module base_cstret_mod
  use  base_struct_mod
  implicit none
  type(axis)                        , private :: am
  type(axis)                        , private :: ad       
  real                              , private :: eps
  integer, dimension(:), allocatable, private :: indx
  logical, dimension(:), allocatable, private :: flag
  real   , dimension(:), allocatable, private :: wght
  real   , dimension(:), allocatable, private :: diag, offd
  real   , dimension(:), allocatable, private :: d, o
  logical                           , private :: push
contains
  
  integer function base_cstret_init(am_in, ad_in, eps_in, push_in) result(stat)
    type(axis) :: am_in, ad_in
    real       :: eps_in
    logical    :: push_in
    stat = 1
    am   = am_in
    ad   = ad_in
    eps  = eps_in
    push = push_in

    if (push) then
       if(.not.allocated(indx)) allocate(indx(ad%n))
       if(.not.allocated(flag)) allocate(flag(ad%n))
       if(.not.allocated(wght)) allocate(wght(ad%n))
    else
       if(.not.allocated(indx)) allocate(indx(am%n))
       if(.not.allocated(flag)) allocate(flag(am%n))
       if(.not.allocated(wght)) allocate(wght(am%n))
    end if
    if(.not.allocated(diag)) allocate(diag(am%n))
    if(.not.allocated(offd)) allocate(offd(am%n))
    if(.not.allocated(d))    allocate(d(am%n))
    if(.not.allocated(o))    allocate(o(am%n))
    stat = 0
  end function

  integer function base_cstret_clos() result(stat)
    stat = 1
    deallocate(indx, flag, wght, diag, offd, d, o)
    stat = 0
  end function
  
  integer function base_cstret_oper(coord, modl, data) result(stat)
    real   , dimension(:) :: coord
    complex, dimension(:) :: modl
    complex, dimension(:) :: data

    integer               :: im, id, il, iu, k
    real                  :: rd, rm, wl, wu
    complex               :: t

    stat = 1

    diag = 0.
    offd = 0.
    d    = 0.
    o    = 0.
    if (push) then
       !! for every data point
       !! - find the nearest lower index in the model   (indx)
       !! - find the weights to the lower/upper indexes (wght)

       do id=1, ad%n
          rd = 1+(coord(id)-am%o)/am%d

          il = rd     !! lower index
          wu = rd-il  !! upper weight (0 if the point is on the grid)

          !! ignore point if out of the grid
          if (il<1 .or. il>am%n-1) then
             flag(id) = .true.
             cycle
          end if

          !! for every point in the data, keep:
          flag(id) = .false.  !! an 'in bounds' flag
          indx(id) = il       !! the lower index
          wght(id) = wu       !! the upper weight

          iu = il+1
          wl = 1.-wu

          !! create tridiagonal (LtL)
          diag(il) = diag(il) + wl*wl
          diag(iu) = diag(iu) + wu*wu
          offd(il) = offd(il) + wl*wu

       end do

    else
       !! for every model point
       !! - find the nearest lower index in the data
       !! - find the weights to the lower/upper indexes (wght)
       
       do im=1, am%n
          rm = 1+(coord(im)-ad%o)/ad%d

          il = rm     !! lower index
          wu = rm-il  !! upper weight (0 if the point is on the grid)

          !! ignore point if out of the grid
          if (il<1 .or. il>ad%n-1) then
             flag(im) = .true.
             cycle
          end if

          !! for every point in the model, keep:
          flag(im) = .false.   !! an 'in bounds' flag
          indx(im) = il        !! the lower index
          wght(im) = wu        !! the upper weight

          iu = il+1
          wl = 1.-wu

          !! create tridiagonal (LtL)
          diag(im) = diag(im) + wl*wl
          diag(im) = diag(im) + wu*wu

       end do

    end if

    !! add regularization: LtL + eps AtA
    diag = diag + 2.*eps
    offd = offd -    eps 

    !!.. stretch apply   
    !!.. compute m = Lt d

    if (push) then
       do id=1, ad%n
          if (flag(id)) cycle

          il = indx(id) !! lower index
          wu = wght(id) !! upper weight

          iu = il+1     !! upper index
          wl = 1.-wu    !! lower weight

          modl(il) = modl(il) + wl*data(id)
          modl(iu) = modl(iu) + wu*data(id)
       end do
    else
       do im=1, am%n
          if (flag(im)) cycle

          il = indx(im)  !! lower index
          wu = wght(im)  !! upper weight

          iu = il+1      !! upper index
          wl = 1.-wu     !! lower weight

          modl(im) = modl(im) + wl*data(il)
          modl(im) = modl(im) + wu*data(iu)

       end do

    end if

    !!.. tridiagonal solver
    !!.. compute (L'L+e^2 A'A) m = L' d
    d(1) = diag(1)
    do k=2, am%n
       t      = offd(k-1) 
       o(k-1) = t/diag(k-1)
       d(k)   = diag(k)-t*o(k-1)
    end do

    do k=2, am%n
       modl(k) = modl(k)-o(k-1)*modl(k-1)
    end do

    modl(am%n) = modl(am%n)/d(am%n)
    do k=am%n-1, 1, -1
       modl(k) = modl(k)/d(k)-o(k)*modl(k+1)
    end do

    stat = 0
  end function

end module
