!!!   Changed by Yi Shen, May 5, 2012
!!!   only changed tayexp, not fourie
!!!   phsscr_tayexp
!!!  Changed by Yi Shen, May 14, 2012
!!!  bwi_bornsc_init, bwi_bornsc_oper: definations

module bwi_bornsc_mod
  use  base_fftwv3_mod
  implicit none

  real, dimension(:), pointer, private :: kx2, ky2
  real                       , private :: dx, dy, dz
  real                       , private :: cc(4)
  integer                    , private :: nx, ny, nvrf, bscnop, nqrf
  character(len=6)           , private :: bsctyp, bsccoe

contains
  
  integer function bwi_bornsc_init(kx2_in, ky2_in, dx_in, dy_in, dz_in, nx_in, ny_in, &
                                   nvrf_in,nqrf_in, bscnop_in, bsctyp_in, bsccoe_in) result(stat)
    real, dimension(:), target :: kx2_in, ky2_in
    real                       :: dx_in, dy_in, dz_in
    integer                    :: nx_in, ny_in, nvrf_in, bscnop_in,  nqrf_in
    character(len=6)           :: bsctyp_in, bsccoe_in

    kx2    => kx2_in
    ky2    => ky2_in
    dx     =  dx_in
    dy     =  dy_in
    dz     =  dz_in
    nx     =  nx_in
    ny     =  ny_in
    nvrf   =  nvrf_in
    nqrf   =  nqrf_in
    bscnop =  bscnop_in
    bsctyp =  bsctyp_in
    bsccoe =  bsccoe_in

    call bwi_phsscr_coef(cc, bsccoe, bscnop)

    stat = 0
  end function

  integer function bwi_bornsc_oper(gwfd, bwfd, wtmp, bvel, bq, vref,qref, vmap, ws, opsign, plan_2dxF, plan_2dxI) result(stat)
    complex, dimension(:,:) :: gwfd
    complex, dimension(:,:) :: bwfd
    complex, dimension(:,:) :: wtmp
    real   , dimension(:,:) :: bvel
    real   , dimension(:)   :: vref
    real   , dimension(:,:) :: bq
    real   , dimension(:)   :: qref
    integer, dimension(:,:) :: vmap
    real                    :: ws
    integer                 :: opsign
    integer*8               :: plan_2dxF, plan_2dxI

    select case(bsctyp)
       case ("fourie")
            call phsscr_fourie(gwfd, bwfd, wtmp, bvel, vref, vmap, ws, opsign, plan_2dxF, plan_2dxI)
       case ("tayexp")
            call phsscr_tayexp(gwfd, bwfd, wtmp, bvel, bq, ws, opsign, plan_2dxF, plan_2dxI)
       case default
            write(0,*)"bsctyp has to be either <fourie> or <tayexp>"
            STOP
    end select

    stat = 0
  end function

  integer function bwi_bornsc_clos() result(stat)
    stat = 1
    kx2 => null()
    ky2 => null()
    dx = 0; dy = 0; dz = 0
    nx = 0; ny = 0
    bscnop = 0; bsctyp = " "; bsccoe=" "
    stat = 0
  end function



! input background wfld, output scattered wfld 
  subroutine phsscr_tayexp(gwfd, bwfd, wtmp, bvel,bq, ws, opsign, plan_2dxF, plan_2dxI)  !!! bvel = true vel?
    complex, dimension(:,:) :: gwfd
    complex, dimension(:,:) :: bwfd
    complex, dimension(:,:) :: wtmp
    real   , dimension(:,:) :: bvel
    real   , dimension(:,:) :: bq
    real                    :: ws
    integer                 :: opsign
    integer*8               :: plan_2dxF, plan_2dxI

    integer                 :: iop, ikx, iky, ix, iy
    complex                 :: isdz
    real                    :: kk2
    complex*16		    :: vel, wov, wov2

!   optimizing in velocity
!   i * sign * dKz/dq * dz
!   dKz/dv = -w/v/v / sqrt(1-v*v/w/w*|k|^2)*(1+i/2q)^(-2)*iv0/2/Q/Q  (Q)
!   -> dKz/dv = -w/v/v / sqrt(1-v*v/w/w*|k|^2)*(1+i/2q)^(-2)*iv0/2*ln(10)*10**(-q)
!   ---------------------------------------

!   set up the sign
!   ---------------
    isdz = cmplx(0., 1.)*dz*opsign

!   compute the expansion
!   1/sqrt(1-s^2) = 1 + c1 * s^2 + c2* s^4 + c3 * s^6 + c4 * s^8 + ...
!   where s = v/w*|k|
!   -----------------------------------------------------------------
    gwfd = bwfd
    do iop=1, bscnop
       wtmp = bwfd 
       call sfftw_dft_2dx_oper(plan_2dxF, wtmp, nx, ny)
       do iky=1, ny
       do ikx=1, nx
          kk2 = kx2(ikx)+ky2(iky)
          wtmp(ikx,iky) = cc(iop)*(kk2**iop)*wtmp(ikx,iky)
       end do
       end do
       call sfftw_dft_2dx_oper(plan_2dxI, wtmp, nx, ny)
       do iy=1, ny
       do ix=1, nx
          vel  = bvel(ix,iy)/(1.+cmplx(0., 1.)/(10.**bq(ix,iy))/2.)  !!! conjugate; no phase change
          wov  = ws/vel
          wov2 = wov*wov
          gwfd(ix,iy) = gwfd(ix,iy) + wtmp(ix,iy)*dble((1./wov2)**iop)
       end do
       end do
    end do
!   multiple the weighting function
!   -------------------------------
    do iy=1, ny
    do ix=1, nx
       vel =bvel(ix,iy)/(1.+cmplx(0., 1.)/(10.**bq(ix,iy))/2.)   !!! conjugate; no phase change
       wov = ws/vel
       gwfd(ix,iy) = gwfd(ix,iy)*isdz*(-wov)/vel/((1.+cmplx(0., 1.)/2./(10.**bq(ix,iy)))**2)*cmplx(0., 1.)*bvel(ix,iy)/2.*log(10.)*(10.**(-bq(ix,iy)))  !!! conjugate
    end do
    end do

    return
  end subroutine

  subroutine phsscr_fourie(gwfd, bwfd, wtmp, bvel, vref, vmap, ws, opsign, plan_2dxF, plan_2dxI)
    complex, dimension(:,:) :: gwfd
    complex, dimension(:,:) :: bwfd
    complex, dimension(:,:) :: wtmp
    real   , dimension(:,:) :: bvel
    real   , dimension(:)   :: vref
    integer, dimension(:,:) :: vmap
    real                    :: ws
    integer                 :: opsign
    integer*8               :: plan_2dxF, plan_2dxI

    complex                 :: isdz
    real                    :: kk2, kz2, kz2max, vdif1, vdif2
    integer                 :: ix, iy, ikx, iky, ivrf, jvrf

    wtmp = 0.
    vmap = 0
!   compute the velocity map
!   ------------------------
    do iy=1, ny
    do ix=1, nx
       jvrf = 1
       do ivrf=1, nvrf-1
          vdif1 = abs(bvel(ix,iy)-vref(ivrf  ))
          vdif2 = abs(bvel(ix,iy)-vref(ivrf+1))
          if (vdif1<vdif2) then
             jvrf = ivrf
          else
             jvrf = ivrf+1
          end if
       end do
       vmap(ix,iy) = jvrf
    end do
    end do

!   set the maximum angle
!   ---------------------
    kz2max = cos(80*3.14159265359/180.)**2

    isdz = cmplx(0., 1.)*dz*opsign

    do ivrf=1, nvrf
       wtmp = bwfd
       call sfftw_dft_2dx_oper(plan_2dxF, wtmp, nx, ny) 
       do iky=1, ny
       do ikx=1, nx
          kk2 = kx2(ikx)+ky2(iky)
          kz2 = 1 - kk2/ws/ws*vref(ivrf)*vref(ivrf)
          if (kz2>=kz2max) then
             wtmp(ikx,iky) = wtmp(ikx,iky)/sqrt(kz2) 
          else 
             wtmp(ikx,iky) = 0.
          end if
       end do
       end do
       call sfftw_dft_2dx_oper(plan_2dxI, wtmp, nx, ny)
       where (vmap == ivrf)
             gwfd = wtmp * (-ws/vref(ivrf)/vref(ivrf))
       end where 
    end do 
    gwfd = gwfd * isdz !* (-ws/bvel/bvel)
    return
  end subroutine

  subroutine bwi_phsscr_coef(cc, bsccoe, bscnop)
    real, dimension(4) :: cc
    character(len=6)   :: bsccoe
    integer            :: bscnop
    select case (bsccoe)
      case ("taylor")
           cc(1) = 1./2.
           cc(2) = 3./8.
           cc(3) = 5./16.
           cc(4) = 35./128.
           if (bscnop>4) then
              write(0,*) "ERROR: bscnop <= 4, Taylor coef"
              STOP
           end if
      case ("optimz")
           select case (bscnop)
             case (2) !! up to 45 degree
                  cc(1) = 0.437271673; cc(2) = 0.743598856
             case (3) !! up to 65 degree
                  cc(1) = 0.646864667; cc(2) = -0.650508335; cc(3) = 2.020601118
             case default
                  write(0,*) "ERROR: 2 =< bscnop <= 3, Optimized coef" 
                  STOP
           end select 
      case default 
           write(0,*)"bsccoe has to be either <taylor> or <optimz>"
           STOP
    end select 
    return
  end subroutine




end module
