module bwi_dipcal_mod
  use base_struct_mod
  use base_trismt_mod
  implicit none

  type(axis)                       , private :: ax, az
  integer                          , private :: rectx, rectz, nsmt
  real, dimension(:,:), allocatable, private :: t
  real, dimension(:)  , allocatable, private :: d, e
  real, dimension(:,:), allocatable, private :: t11, t12, t22

contains

  integer function bwi_dipcal_init(ax_in, az_in, rectx_in, rectz_in, nsmt_in) result(stat)
    type(axis) :: ax_in, az_in
    integer    :: rectx_in, rectz_in, nsmt_in

    integer    :: nd

    ax    = ax_in
    az    = az_in
    rectx = rectx_in
    rectz = rectz_in
    nsmt  = nsmt_in
    allocate(t11(ax%n,az%n))
    allocate(t12(ax%n,az%n))
    allocate(t22(ax%n,az%n))
    allocate(t(2,2), d(2), e(2))
    stat = 0
  end function

  integer function bwi_dipcal_clos() result(stat)
    stat = 1
    deallocate(t, d, e)
    deallocate(t11, t12, t22)
    stat = 0
  end function

  integer function bwi_dipcal_oper(f, v2) result(stat)
    real, dimension(:,:) :: f, v2 ! v2 is the dip [-pi/2, pi/2]

    integer              :: ix, iz
    real                 :: fx, fz, gx, gz
    real                 :: PI

    PI = 3.14159265359 

    t = 0.; d = 0.; e = 0.

!   2-D filter
!   ----------
    t11 = 0.
    t12 = 0.
    t22 = 0.
    v2  = 0.
    do iz=2, az%n
    do ix=2, ax%n
!      compute the gradient component
!      ------------------------------
!       fx = 0.5*(f(ix,iz)-f(ix-1,iz))/ax%d + 0.5*(f(ix,iz-1)-f(ix-1,iz-1))/ax%d
!       fz = 0.5*(f(ix,iz)-f(ix,iz-1))/az%d + 0.5*(f(ix-1,iz)-f(ix-1,iz-1))/az%d
       fx = (f(ix,iz)-f(ix-1,iz))/ax%d
       fz = (f(ix,iz)-f(ix,iz-1))/az%d 

!      form the structure tensor
!      -------------------------
       t11(ix,iz) = fx*fx 
       t12(ix,iz) = fx*fz
       t22(ix,iz) = fz*fz
    end do
    end do
!   smooth the structure tensor
    stat = base_tris2d_oper(t11,rectx,rectz,nsmt)
    stat = base_tris2d_oper(t12,rectx,rectz,nsmt)
    stat = base_tris2d_oper(t22,rectx,rectz,nsmt) 
!   perform eigen decomposition
    do iz=min(3,az%n), max(1,az%n-3)
    do ix=min(3,az%n), max(1,ax%n-3)
       t(1,1) = t11(ix,iz)
       t(1,2) = t12(ix,iz)
       t(2,1) = t12(ix,iz)
       t(2,2) = t22(ix,iz)
       d = 0.
       e = 0. 
       call tred2(t,2,2,d,e)
       call tqli (d,e,2,2,t)

       if (abs(d(1))>abs(d(2))) then
!          u1(ix,1,iz) = t(1,1); u2(ix,1,iz) = t(2,1)
!          v1(ix,1,iz) = t(1,2); v2(ix,1,iz) = t(2,2) 
          v2(ix,iz) = t(2,2)
       else 
!          u1(ix,1,iz) = t(1,2); u2(ix,1,iz) = t(2,2)
!          v1(ix,1,iz) = t(1,1); v2(ix,1,iz) = t(2,1)
          v2(ix,iz) = t(2,1)
       end if 
    end do
    end do
    stat = base_tris2d_oper(v2,2,2,1)
    v2 = asin(v2)*180./PI
    stat = 0
  end function

! Householder reduction of a real, symmetric, n by n matrix a, 
! stored in an np by np physical array. On output, a is replaced
! by the orthogonal matrix Q effecting the transformation. d
! returns the diagonal elements of the tridiagonal matrix, and
! e the off-diagonal elements, with e(1)=0. Several statements,
! as noted in comments, can be omitted if only eigenvalues are
! to be found, in which case a contains no useful information
! on output. Otherwise they are to be included.
  subroutine tred2(a,n,np,d,e)
    integer :: n,np
    real    :: a(np,np),d(np),e(np)
    integer :: i,j,k,l
    real    :: f,g,h,hh,scale
    do i=n,2,-1
      l=i-1
      h=0.
      scale=0.
      if(l.gt.1)then
        do k=1,l
           scale=scale+abs(a(i,k))
        end do
        if(scale.eq.0.)then
          e(i)=a(i,l)
        else
          do k=1,l
             a(i,k)=a(i,k)/scale
             h=h+a(i,k)**2
          end do
          f=a(i,l)
          g=-sign(sqrt(h),f)
          e(i)=scale*g
          h=h-f*g
          a(i,l)=f-g
          f=0.
          do j=1,l
!     Omit following line if finding only eigenvalues
            a(j,i)=a(i,j)/h
            g=0.
            do k=1,j
              g=g+a(j,k)*a(i,k)
            end do
            do k=j+1,l
              g=g+a(k,j)*a(i,k)
            end do
            e(j)=g/h
            f=f+e(j)*a(i,j)
          end do
          hh=f/(h+h)
          do j=1,l
            f=a(i,j)
            g=e(j)-hh*f
            e(j)=g
            do k=1,j
              a(j,k)=a(j,k)-f*e(k)-g*a(i,k)
            end do
          end do
        endif
      else
        e(i)=a(i,l)
      endif
      d(i)=h
    end do
!   Omit following line if finding only eigenvalues.
    d(1)=0.
    e(1)=0.
    do i=1,n
!   Delete lines from here ...
      l=i-1
      if(d(i).ne.0.)then
        do j=1,l
          g=0.
          do k=1,l
            g=g+a(i,k)*a(k,j)
          end do
          do k=1,l
            a(k,j)=a(k,j)-g*a(k,i)
          end do
        end do
      endif
!   ... to here when finding only eigenvalues.
      d(i)=a(i,i)
!   Also delete lines from here ...
      a(i,i)=1.
      do j=1,l
        a(i,j)=0.
        a(j,i)=0.
      end do
!   ... to here when finding only eigenvalues.
    end do
    return
  end subroutine

! QL algorithm with implicit shifts, to determine the eigenvalues and
! eigenvectors of a real, symmetric, tridiagonal matrix, or of a real,
! symmetric matrix previously reduced by tred2. d is a vector of lenght
! np. On input, its first n elements are the diagonal elements of the 
! tridiagonal matrix. On output, it returns the eigenvalues. The vector
! e is destroyed. When finding only the eigenvalues, several lines may 
! be omitted, as noted in the comments. If the eigenvectors of a 
! tridiagonal matrix are desired, the matrix z (n by n matrix stored in
! np by np array) is input as the identity matrix. If the eigenvectors
! of a matrix that has been reduced by tred2 are required, then z is 
! input as the matrix output by tred2. In either case, the kth column
! of z returns the normalized eigenvector corresponding to d(k).
  subroutine tqli(d,e,n,np,z)
    integer :: n,np
    real    :: d(np),e(np),z(np,np)
    integer :: i,iter,k,l,m
    real    :: b,c,dd,f,g,p,r,s
    do i=2,n
       e(i-1)=e(i)
    end do
    e(n)=0.
    do l=1,n
      iter=0
1     do m=l,n-1
         dd=abs(d(m))+abs(d(m+1))
         if (abs(e(m))+dd.eq.dd) goto 2
      end do
      m=n
2     if(m.ne.l)then
        if (iter.eq.100) then
!           d = 0.
!           z = 1.
           return
        end if
        iter=iter+1
        g=(d(l+1)-d(l))/(2.*e(l))
        r=pythag(g,1.)
        g=d(m)-d(l)+e(l)/(g+sign(r,g))
        s=1.
        c=1.
        p=0.
        do i=m-1,l,-1
           f=s*e(i)
           b=c*e(i)
           r=pythag(f,g)
           e(i+1)=r
           if(r.eq.0.)then
             d(i+1)=d(i+1)-p
             e(m)=0.
             goto 1
           endif
           s=f/r
           c=g/r
           g=d(i+1)-p
           r=(d(i)-g)*s+2.*c*b
           p=s*r
           d(i+1)=g+p
           g=c*r-b
!   Omit lines from here ...
           do k=1,n
              f=z(k,i+1)
              z(k,i+1)=s*z(k,i)+c*f
              z(k,i)=c*z(k,i)-s*f
           end do
!   ... to here when finding only eigenvalues.
        end do
        d(l)=d(l)-p
        e(l)=g
        e(m)=0.
        goto 1
      endif
    end do
    return
  end subroutine

  function pythag(a,b)
    real :: a,b,pythag
    real :: absa,absb
    absa=abs(a)
    absb=abs(b)
    if(absa.gt.absb)then
      pythag=absa*sqrt(1.+(absb/absa)**2)
    else
      if(absb.eq.0.)then
        pythag=0.
      else
        pythag=absb*sqrt(1.+(absa/absb)**2)
      endif
    endif
    return
  end function 


end module
