!!.. this module contains various kinds of one-way extrapolators
!!--------------------------------------------------------------------
!!.. Author: Yaxun Tang, tang@sep.stanford.edu
!!--------------------------------------------------------------------
!!   Add Q in phs, ssf, vel_correct, q_correct --Yi Shen (Nov 28, 2011)
!!   create vtilde to follow the original routine instead of using q_correct, only change vel_correct--Yi Shen (Dec 5,2011)
!!   Add Q in phase shift, shift_correct,phs, ssf and delete vel_correct, q_correct---Yi Shen (Jan 18, 2012) 
!!   Add Q in ssr_ffd (calling and defination only to prevent error report)  -- Yi Shen (Jan 19, 2012)
!!   psign -> opsign in phase_shift and phase_correct  -- Yi Shen (Feb 29, 2012) (adj in ssr_ssf:Potential bug?)--change back by Yi Shen (Apr 5, 2012)
!!--------------------------------------------------------------------
!!   Add Q compensation -Apr 5, 2012
!!	1) ssr_phs: define, phase_shift
!!	2) ssr_ssf: define, phase_shift
!!	3) phase_shift
!!	4) ssr_ffd: phase_shift(calling and defination only to prevent error report)
!!	5) Notice: ssf_correct needs change?
!!   Change hard clip to angle limitation - Jul 16, 2013 (BP)
module bwi_extrpl_mod
  use  base_fftwv3_mod
  implicit none

contains
  !!---------------------------------------------------------------------------------------------------------
  !!.. Phase-shift extrapolator with Q 
  !!---------------------------------------------------------------------------------------------------------
  integer function ssr_phs(adj, wfld_slice, vmod_slice, qmod_slice,kx2, ky2, vref, qref, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI, qcomp, qcomp_clip,p1, p2, p3 ,p4) result(stat)
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:,:) :: vmod_slice
    real   , dimension(:)   :: kx2, ky2

    real                    :: vref
    real                    :: ws
    real                    :: dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign
    integer*8               :: plan_2dxF
    integer*8               :: plan_2dxI
    real		    :: qref            !   Add q
    real   , dimension(:,:) :: qmod_slice      !   Add qmod

    logical	    	    :: qcomp
    real	    	    :: qcomp_clip
    real		    :: p1, p2, p3 ,p4

    stat = 1

    if (adj) then
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
       call phase_shift(adj, wfld_slice, kx2, ky2, vref,qref, ws, dz, nx, ny, opsign, qcomp, qcomp_clip,p1, p2, p3 ,p4)!!!Add qref
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
    else
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny) 
       call phase_shift(adj, wfld_slice, kx2, ky2, vref,qref, ws, dz, nx, ny, opsign, qcomp, qcomp_clip,p1, p2, p3 ,p4) !!! Add qref
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
    end if
    stat = 0
  end function

  !!---------------------------------------------------------------------------------------------------------
  !!.. Splt-step Fourier extrapolator with Q
  !!---------------------------------------------------------------------------------------------------------
  integer function ssr_ssf(adj, wfld_slice, vmod_slice, qmod_slice, kx2, ky2, vref, qref, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI, qcomp, qcomp_clip,p1,p2,p3,p4) result(stat)  !!!Add qmod_slice, qref
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:,:) :: vmod_slice
    real   , dimension(:)   :: kx2, ky2
    real                    :: vref
    real                    :: ws
    real                    :: dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign
    integer*8               :: plan_2dxF
    integer*8               :: plan_2dxI
    real		    :: qref            !   Add q
    real   , dimension(:,:) :: qmod_slice      !   Add qmod
    logical	    	    :: qcomp
    real	    	    :: qcomp_clip
   real 		    :: p1, p2, p3, p4,t1,t2

    stat = 1
    if (adj) then
! 	call cpu_time ( t1 )
       call ssf_correct(adj, wfld_slice, vmod_slice, qmod_slice, vref,qref, ws, dz, nx, ny, opsign)      !!! Add qmod_slice, qref
! 	call cpu_time ( t2 )
! 	p1=t2-t1+p1
! 	call cpu_time ( t1 )
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
! 	call cpu_time ( t2 )
! 	p2=t2-t1+p2
! 	call cpu_time ( t1 )
       call phase_shift(adj, wfld_slice, kx2, ky2, vref,qref,  ws, dz, nx, ny, opsign, qcomp, qcomp_clip,p1, p2, p3 ,p4)			!!! Add qref
! 	call cpu_time ( t2 )
! 	p3=t2-t1+p3
! 	call cpu_time ( t1 )
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
! 	call cpu_time ( t2 )
! 	p4=t2-t1+p4
    else 

! 	call cpu_time ( t1 )
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
! 	call cpu_time ( t2 )
! 	p1=t2-t1+p1
!	write(0,*) "1",t2,t1,t2-t1
! 	call cpu_time ( t1 )
       call phase_shift(adj, wfld_slice, kx2, ky2, vref, qref, ws, dz, nx, ny, opsign, qcomp, qcomp_clip,p1, p2, p3 ,p4)			!!! Add qref
! 	call cpu_time ( t2 )
! 	p2=t2-t1+p2
!	write(0,*) "2",t2,t1,t2-t1
! 	call cpu_time ( t1 )
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
! 	call cpu_time ( t2 )
! 	p3=t2-t1+p3
!	write(0,*) "3",t2,t1,t2-t1
! 	call cpu_time ( t1 )
       call ssf_correct(adj, wfld_slice, vmod_slice, qmod_slice, vref, qref, ws, dz, nx, ny, opsign)	!!! Add qmod_slice,qref
! 	call cpu_time ( t2 )
! 	p4=t2-t1+p4
!	write(0,*) "4",t2,t1,t2-t1

    end if
    stat = 0

  end function
  
  !!-----------------------------------------------------------------------------------------------------------------------
  !!.. Fourier finite-difference extrapolator
  !!-----------------------------------------------------------------------------------------------------------------------
  integer function ssr_ffd(adj, wfld_slice, vmod_slice, kx2, ky2, vref, ws, dx, dy, dz, trick, bndp, &
                           nx, ny, dip, opsign, bullet, rwrk1d, cwrk1d, plan_2dxF, plan_2dxI) result(stat)  
    logical                          :: adj
    complex, dimension(:,:)          :: wfld_slice
    real   , dimension(:,:)          :: vmod_slice
    real   , dimension(:)            :: kx2, ky2
    real                             :: vref
    real                             :: ws
    real                             :: dx, dy, dz, trick
    complex                          :: bndp
    integer                          :: nx, ny, dip
    integer, intent(in)              :: opsign
    logical                          :: bullet
    real   , dimension(nx+ny)        :: rwrk1d
    complex, dimension(nx*5+ny*5)    :: cwrk1d
    integer*8                        :: plan_2dxF
    integer*8                        :: plan_2dxI

    real                             :: alph(5), beta(5), mmax
    integer                          :: nstep
    real		    	     :: qref            !   Add q
    real   , dimension(nx,ny) 	     :: qmod_slice      !   Add qmod
    logical	    	 	     :: qcomp
    real	    	       	     :: qcomp_clip
   real					:: p1, p2, p3 ,p4
    stat = 1


    mmax = maxval(vmod_slice)/vref

    call ffd_coeffts(alph, beta, mmax, dip)

    if (adj) then
       call ffd_correct(adj, wfld_slice, vmod_slice, alph, beta, vref, ws, dx, dy, dz, trick, bndp, &
                        nx, ny, nstep, dip, opsign, bullet, rwrk1d, cwrk1d)
       call ssf_correct(adj, wfld_slice, vmod_slice, qmod_slice, vref, qref, ws, dz, nx, ny, opsign)        !!!Add qmod_slice, qref
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
       call phase_shift(adj, wfld_slice, kx2, ky2, vref, qref, ws, dz, nx, ny, opsign, qcomp, qcomp_clip,p1, p2, p3 ,p4)           		!!! Add qref
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny) 
    else
       call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
       call phase_shift(adj, wfld_slice, kx2, ky2, vref, qref, ws, dz, nx, ny, opsign,qcomp, qcomp_clip,p1, p2, p3 ,p4)          		!!! Add qref
       call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
       call ssf_correct(adj, wfld_slice, vmod_slice, qmod_slice, vref, qref, ws, dz, nx, ny, opsign)     !!!Add qmod_slice, qref
       call ffd_correct(adj, wfld_slice, vmod_slice, alph, beta, vref, ws, dx, dy, dz, trick, bndp, &
                        nx, ny, nstep, dip, opsign, bullet, rwrk1d, cwrk1d)
    end if
    stat = 0
  end function

!  integer function ssr_efd() result(stat)
!  end function

  !!-----------------------------------------------------------------------------------------------------------
  !!.. Frequency domain implicit finite-difference extrapolator
  !!.. Dip upto 15, 45, 65, 79, 80, 87, 89, 90
  !!-----------------------------------------------------------------------------------------------------------
  integer function ssr_ifd(adj, wfld_slice, vmod_slice, kx2, ky2, vref, ws, dx, dy, dz, trick, bndp, &
                           nx, ny, dip, opsign, bullet, errflt, rwrk1d, cwrk1d, plan_2dxF, plan_2dxI) result(stat)
    logical                          :: adj
    complex, dimension(:,:)          :: wfld_slice
    real   , dimension(:,:)          :: vmod_slice
    real   , dimension(:)            :: kx2, ky2
    real                             :: vref, ws
    real                             :: dx, dy, dz, trick
    complex                          :: bndp
    integer                          :: nx, ny, dip
    integer, intent(in)              :: opsign
    logical                          :: bullet, errflt
    real   , dimension(nx+ny)        :: rwrk1d
    complex, dimension(nx*5+ny*5)    :: cwrk1d

    integer*8                        :: plan_2dxF
    integer*8                        :: plan_2dxI

    integer                          :: ix, iy, istep, nstep
    integer                          :: lrx1, lry1
    integer                          :: lcx1, lcx2, lcx3, lcx4, lcx5
    integer                          :: lcy1, lcy2, lcy3, lcy4, lcy5
    real                             :: alph(5), beta(5)

    !!.. asign work space
    lcx1 = 1            !!.. wfdx
    lcx2 = lcx1 + nx    !!.. ahx
    lcx3 = lcx2 + nx    !!.. bhx
    lcx4 = lcx3 + nx    !!.. chx
    lcx5 = lcx4 + nx    !!.. rhx
    lcy1 = lcx5 + nx    !!.. wfdy
    lcy2 = lcy1 + ny    !!.. ahy
    lcy3 = lcy2 + ny    !!.. bhy
    lcy4 = lcy3 + ny    !!.. chy
    lcy5 = lcy4 + ny    !!.. rhy
    
    lrx1 = 1            !!.. velx
    lry1 = lrx1 + nx    !!.. vely

    call ifd_coeffts(alph, beta, nstep, dip)

    if (adj) then

!      apply error filter if requested
!      -------------------------------
       if (errflt) then
          call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
          call ifd_errfltr(adj, wfld_slice, kx2, ky2, alph, beta, trick, ws, vref, &
                           dx, dy, dz, nx, ny, nstep, opsign)
          call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
       end if

!      solver for diffraction term in y first
!     ---------------------------------------
       if (ny > 1) then
          do ix=1, nx
             cwrk1d(lcy1:lcy2-1) = wfld_slice(ix,:)
             rwrk1d(lry1:nx+ny ) = vmod_slice(ix,:)
             do istep=nstep, 1, -1
                call ifd_sysform(adj, cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), rwrk1d(lry1), ws, dy, dz, &
                                 alph(istep), beta(istep), trick, bndp, ny, opsign, bullet)
                call ifd_dfrterm(adj, cwrk1d(lcy1), cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), cwrk1d(lcy5), ny)
             end do
             wfld_slice(ix,:) = cwrk1d(lcy1:lcy2-1)
          end do
       end if

!      solve for diffraction term in x
!      ------------------------------- 
       if (nx > 1) then
          do iy=1, ny
             cwrk1d(lcx1:lcx2-1) = wfld_slice(:,iy)
             rwrk1d(lrx1:lry1-1) = vmod_slice(:,iy)
             do istep=nstep, 1, -1
                call ifd_sysform(adj, cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), rwrk1d(lrx1), ws, dx, dz, &
                                 alph(istep), beta(istep), trick, bndp, nx, opsign, bullet)
                call ifd_dfrterm(adj, cwrk1d(lcx1), cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), cwrk1d(lcx5), nx)
             end do
             wfld_slice(:,iy) = cwrk1d(lcx1:lcx2-1)
          end do
       end if

!      solve for the thin lens term
!      ----------------------------
       call ifd_sftterm(adj, wfld_slice, vmod_slice, ws, dz, nx, ny, opsign)

    else

!      solve for the thin lens term
!      ----------------------------
       call ifd_sftterm(adj, wfld_slice, vmod_slice, ws, dz, nx, ny, opsign)

!      solve for the diffraction term in x first
!      -----------------------------------------
       if (nx > 1) then
          do iy=1, ny
             cwrk1d(lcx1:lcx2-1) = wfld_slice(:,iy)
             rwrk1d(lrx1:lry1-1) = vmod_slice(:,iy)
             do istep=1, nstep, 1
                call ifd_sysform(adj, cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), rwrk1d(lrx1), ws, dx, dz, &
                                 alph(istep), beta(istep), trick, bndp, nx, opsign, bullet)
                call ifd_dfrterm(adj, cwrk1d(lcx1), cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), cwrk1d(lcx5), nx)
             end do
             wfld_slice(:,iy) = cwrk1d(lcx1:lcx2-1)
          end do
       end if

!      solve for the diffraction term in y
!      ----------------------------------- 
       if (ny > 1) then
          do ix=1, nx
             cwrk1d(lcy1:lcy2-1) = wfld_slice(ix,:)
             rwrk1d(lry1:nx+ny ) = vmod_slice(ix,:)
             do istep=1, nstep, 1
                call ifd_sysform(adj, cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), rwrk1d(lry1), ws, dy, dz, &
                                 alph(istep), beta(istep), trick, bndp, ny, opsign, bullet)
                call ifd_dfrterm(adj, cwrk1d(lcy1), cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), cwrk1d(lcy5), ny)
             end do
             wfld_slice(ix,:) = cwrk1d(lcy1:lcy2-1)
          end do
       end if

!      apply error filter, if requested
!      --------------------------------
       if (errflt) then
          call sfftw_dft_2dx_oper(plan_2dxF, wfld_slice, nx, ny)
          call ifd_errfltr(adj, wfld_slice, kx2, ky2, alph, beta, trick, ws, vref, &
                           dx, dy, dz, nx, ny, nstep, opsign)
          call sfftw_dft_2dx_oper(plan_2dxI, wfld_slice, nx, ny)
       end if
    end if
    stat = 0
  end function

  !!-------------------------------------------------------------------------------------------
  !!.. solving for diffraction term
  !!------------------------------------------------------------------------------------------- 
  subroutine ifd_dfrterm(adj, wfdh, ah, bh, ch, rh, nh)
    logical                :: adj
    complex, dimension(nh) :: wfdh
    complex, dimension(nh) :: ah, bh, ch, rh
    integer                :: nh

    integer                :: ih

    rh = 0.

    if (adj) then
       call ctridiag(ah,bh,ch,wfdh,rh,nh)
       do ih=2, nh-1
          wfdh(ih) = conjg(ah(ih))*rh(ih-1) &
                   + conjg(bh(ih))*rh(ih  ) &
                   + conjg(ch(ih))*rh(ih+1)
       end do
       wfdh(1)  = conjg(bh(1))*rh(1) &
                + conjg(ch(1))*rh(2)
       wfdh(nh) = conjg(ah(nh))*rh(nh-1) &
                + conjg(bh(nh))*rh(nh  )

    else
       ! form the right hand side of the equation
       do ih=2, nh-1
          rh(ih) = conjg(ah(ih))*wfdh(ih-1) &
                 + conjg(bh(ih))*wfdh(ih  ) &
                 + conjg(ch(ih))*wfdh(ih+1)
       end do
       ! boundary points
       rh(1)  = conjg(bh(1))*wfdh(1) &
              + conjg(ch(1))*wfdh(2)
       rh(nh) = conjg(ah(nh))*wfdh(nh-1) &
              + conjg(bh(nh))*wfdh(nh  )
       ! tri-diagonal solver
       call ctridiag(ah,bh,ch,rh,wfdh,nh)
    end if
    return
  end subroutine


  !!--------------------------------------------------------------------------
  !!.. solving for shifting term
  !!--------------------------------------------------------------------------
  subroutine ifd_sftterm(adj, wfld_slice, vmod_slice, ws, dz, nx, ny, opsign)
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:,:) :: vmod_slice
    real                    :: ws, dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign

    integer                 :: ix, iy, psign
    real                    :: wov, shift

    if (adj) then
       psign = -opsign
    else
       psign =  opsign
    end if   

    do iy=1, ny
    do ix=1, nx
       wov = ws / vmod_slice(ix,iy) 
       shift = psign * wov * dz 
       wfld_slice(ix,iy) = wfld_slice(ix,iy) * cmplx(cos(shift), sin(shift))
    end do 
    end do

    return
  end subroutine

  !!---------------------------------------------------------------------------
  !!.. finite-difference error compensation filter
  !!.. Li, Z., Compensating finite-difference errors in 3-D migration and 
  !!.. modeling, 1991, VOL. 56, Geophysics. 
  !!---------------------------------------------------------------------------
  subroutine ifd_errfltr(adj, wfld_slice, kx2, ky2, alph, beta, trick, ws, vref, &
                         dx, dy, dz, nx, ny, nstep, opsign)
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:)   :: kx2, ky2
    real   , dimension(5)   :: alph, beta
    real                    :: trick
    real                    :: ws, vref
    real                    :: dx, dy, dz 
    integer                 :: nx, ny, nstep, opsign

    integer                 :: ix, iy, istep
    real                    :: sx, sy, wov, laplx, laply
    real                    :: kz2, kz, phase, kx, ky
    complex                 :: spltx, splty, aax, bbx, aay, bby

    wov = ws / vref 

    if (adj) then
       do iy=1, ny
       do ix=1, nx
          sx = -kx2(ix)/wov/wov
          sy = -ky2(iy)/wov/wov
          kx  = sqrt(kx2(ix))  !! this sign is not important, since cosine function is an even function
          ky  = sqrt(ky2(iy))
          laplx = (2.-2.*cos(kx*dx))/dx/dx
          laply = (2.-2.*cos(ky*dy))/dy/dy
          spltx = cmplx(1.,0.)
          splty = cmplx(1.,0.)
          do istep=1, nstep
!            imaginary coefficients
             aax = opsign*cmplx(0.,1.)*dz*0.5*alph(istep)/wov
             aay = opsign*cmplx(0.,1.)*dz*0.5*alph(istep)/wov
!            real coefficients 
             bbx = trick*dx*dx + beta(istep)/wov/wov
             bby = trick*dy*dy + beta(istep)/wov/wov
!            calculate the dispersion relation for FD scheme
             spltx = spltx*(1.-(bbx+aax)*laplx)/(1.-(bbx-aax)*laplx)
             splty = splty*(1.-(bby+aay)*laply)/(1.-(bby-aay)*laply)
          end do
          kz2 = 1. + sx + sy
          if (kz2>=0.) then
             kz = sqrt(kz2)
             phase = opsign*wov*(kz-1)*dz
             wfld_slice(ix,iy) = wfld_slice(ix,iy)*conjg(cmplx(cos(phase),sin(phase)))
             wfld_slice(ix,iy) = wfld_slice(ix,iy)/conjg(spltx)/conjg(splty)
          else
             kz = sqrt(-kz2)
             phase = -wov*kz*dz
             wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp(phase)
          end if
       end do
       end do

    else
       do iy=1, ny
       do ix=1, nx
          sx = -kx2(ix)/wov/wov
          sy = -ky2(iy)/wov/wov
          kx  = sqrt(kx2(ix))  !! this sign is not important, since cosine function is an even function
          ky  = sqrt(ky2(iy))
          laplx = (2.-2.*cos(kx*dx))/dx/dx
          laply = (2.-2.*cos(ky*dy))/dy/dy
          spltx = cmplx(1.,0.)
          splty = cmplx(1.,0.)
          do istep=1, nstep
!            imaginary coefficients
             aax = opsign*cmplx(0.,1.)*dz*0.5*alph(istep)/wov
             aay = opsign*cmplx(0.,1.)*dz*0.5*alph(istep)/wov
!            real coefficients 
             bbx = trick*dx*dx + beta(istep)/wov/wov
             bby = trick*dy*dy + beta(istep)/wov/wov
!            calculate the dispersion relation for FD scheme
             spltx = spltx*(1.-(bbx+aax)*laplx)/(1.-(bbx-aax)*laplx)
             splty = splty*(1.-(bby+aay)*laply)/(1.-(bby-aay)*laply)
          end do
          kz2 = 1. + sx + sy
          if (kz2>=0.) then
             kz = sqrt(kz2)
             phase = opsign*wov*(kz-1)*dz
             wfld_slice(ix,iy) = wfld_slice(ix,iy)*cmplx(cos(phase),sin(phase))
             wfld_slice(ix,iy) = wfld_slice(ix,iy)/spltx/splty
          else
             kz = sqrt(-kz2)
             phase = -wov*kz*dz
             wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp(phase)
          end if
       end do
       end do
    end if
    return
  end subroutine

  !!---------------------------------------------------------------------------------------
  !!.. Forming the tridiagonal matrix for implicit finite-difference extrapolation
  !!---------------------------------------------------------------------------------------
  subroutine ifd_sysform(adj, ah, bh, ch, velh, ws, dh, dz, alph, beta, trick, bndp, nh, opsign, bullet)
    logical                :: adj
    complex, dimension(nh) :: ah, bh, ch
    real   , dimension(nh) :: velh
    real                   :: ws, dh, dz
    real                   :: alph, beta, trick
    complex                :: bndp
    integer                :: nh, opsign
    logical                :: bullet

    integer                :: ih
    real                   :: velm, velo, velp
    complex                :: aam, aao, aap, bbm, bbo, bbp

    ah = 0; bh = 0; ch = 0

    if (bullet) then

       do ih=1, nh
          velo = velh(ih)
          velm = velh(max(1   , ih-1))
          velp = velh(min(ih+1, nh  ))
!         imaginary coefficients 
          aam  = opsign*cmplx(0.,1.)*dz*0.5*alph/ws*sqrt(velo)*sqrt(velm)/dh/dh
          aao  = opsign*cmplx(0.,1.)*dz*0.5*alph/ws*sqrt(velo)*sqrt(velo)/dh/dh
          aap  = opsign*cmplx(0.,1.)*dz*0.5*alph/ws*sqrt(velo)*sqrt(velp)/dh/dh
!         real coefficients 
          bbm  = trick + beta/ws/ws*velo*velm/dh/dh
          bbo  = trick + beta/ws/ws*velo*velo/dh/dh
          bbp  = trick + beta/ws/ws*velo*velp/dh/dh 
!         sub-diagoal
          ah(ih) = bbm - aam
!         super-diagonal
          ch(ih) = bbp - aap
!         diagonal
          bh(ih) = 1 - 2.*(bbo-aao)
!         boundary condition
          if (ih==1 .or. ih==nh) bh(ih) = 1 - (1+bndp)*(bbo-aao)
       end do

    else

       do ih=1, nh
          velo = velh(ih)
!         imaginary coefficients
          aao  = opsign*cmplx(0.,1.)*dz*0.5*alph/ws*sqrt(velo)*sqrt(velo)/dh/dh
!         real coefficients
          bbo  = trick + beta/ws/ws*velo*velo/dh/dh
!         sub-diagoal
          ah(ih) = bbo - aao
!         super-diagonal
          ch(ih) = bbo - aao
!         diagonal
          bh(ih) = 1 - 2.*(bbo-aao)
!         boundary condition
          if (ih==1 .or. ih==nh) bh(ih) = 1 - (1+bndp)*(bbo-aao)
       end do

    end if  

    if (adj) then
!      transpose and conjugate the matrix
       do ih=1, nh-1
          ah(1)    = ah(ih+1)
          ah(ih+1) = ch(ih)
          ch(ih)   = ah(1)
       end do
       ah = conjg(ah)
       bh = conjg(bh)
       ch = conjg(ch)
    end if

!   not used elements
    ah(1)  = 0.
    ch(nh) = 0.

    return 
  end subroutine

  !!----------------------------------------------------------------------------------------------
  !!.. forming the tridiagonal matrix for FFD extrapolation
  !!----------------------------------------------------------------------------------------------
  subroutine ffd_sysform2(adj, ah, bh, ch, velh, vref, ws, dh, dz, alph, beta, trick, bndp, dip, nh, opsign, bullet)
    logical                :: adj
    complex, dimension(nh) :: ah, bh, ch
    real   , dimension(nh) :: velh
    real                   :: vref, ws, dh, dz
    real                   :: alph, beta, trick
    complex                :: bndp
    integer                :: dip, nh, opsign
    logical                :: bullet

    integer                :: ih
    real                   :: wov, rov, vefc, velo, velm, velp
    real                   :: sum_sq_velo, sum_sq_velm, sum_sq_velp
    complex                :: aah, bbh, aao, bbo, aam, bbm, aap, bbp

    ah = 0; bh = 0; ch = 0
 
    do ih=1, nh
       wov = ws/velh(ih)
       rov = vref/velh(ih)
       aah = opsign * alph*cmplx(0.,1.)*dz/2./wov/dh/dh*(1.-rov)
       bbh = beta/dh/dh/wov/wov*(rov*rov+rov+1.)+trick
       ah(ih) = bbh - aah
       ch(ih) = bbh - aah
       bh(ih) = 1. - 2.*(bbh - aah)
       if (ih==1 .or. ih==nh) bh(ih) = 1. - (1.+bndp)*(bbh-aah)
    end do

    if (adj) then
!      transpose and conjugate the matrix
       do ih=1, nh-1
          ah(1)    = ah(ih+1)
          ah(ih+1) = ch(ih)
          ch(ih)   = ah(1)
       end do
       ah = conjg(ah)
       bh = conjg(bh)
       ch = conjg(ch)
    end if
    ah(1 ) = 0.
    ch(nh) = 0.
    return
  end subroutine

  !!------------------------------------------------------------
  !!..same as ffd_sysform, but slightly slower
  !!------------------------------------------------------------
  subroutine ffd_sysform3(ah, bh, ch, velh, vref, ws, dh, dz, alph, beta, trick, bndp, dip, nh, opsign, bullet)
    complex, dimension(nh) :: ah, bh, ch
    real   , dimension(nh) :: velh
    real                   :: vref, ws, dh, dz
    real                   :: alph, beta, trick
    complex                :: bndp
    integer                :: dip, nh, opsign
    logical                :: bullet

    integer                :: ih
    real                   :: velo, velm, velp, c1o, c1m, c1p, c2o, c2m, c2p, c3o, c3m, c3p
    real                   :: gam, gao, gap, gbm, gbo, gbp
    complex                :: aah, bbh, aao, bbo, aam, bbm, aap, bbp

    ah = 0.; bh = 0.; ch = 0.

vref = minval(velh)

    if (bullet) then
       do ih=1, nh
          velo = velh(ih)
          velm = velh(max(1   ,ih-1))
          velp = velh(min(ih+1,nh  ))

          ! alph[v] = alph * |1-c/v|
          gam = alph*abs(1.-vref/velm)
          gao = alph*abs(1.-vref/velo)
          gap = alph*abs(1.-vref/velp) 

          select case (dip)
            case (45)
                 ! beta[v] = beta*(c*c/v/v+c/v+1)
                 gbm = beta*(vref*vref/velm/velm+vref/velm+1.)
                 gbo = beta*(vref*vref/velo/velo+vref/velo+1.)
                 gbp = beta*(vref*vref/velp/velp+vref/velp+1.)
            case (65)
                 ! beta[v] = beta*(c*c/v/v+1)
                 gbm = beta*(vref*vref/velm/velm+1.)
                 gbo = beta*(vref*vref/velo/velo+1.)
                 gbp = beta*(vref*vref/velp/velp+1.)
            case default
                 write(0,*) "ERROR: dip has to be either 45 or 65 for FFD"
                 stop
          end select

          ! C1 = v*sqrt(beta[v])
!          c1m = velm*sqrt(gbm)
!          c1o = velo*sqrt(gbo)
!          c1p = velp*sqrt(gbp)

c1m = velo*velo*gbo
c1o = velo*velo*gbo+velp*velp*gbp
c1p = velp*velp*gbp

          ! C2 = alph[v]/beta[v]/v
!          c2m = sqrt(gam/gbm/velm)
!          c2o = sqrt(gao/gbo/velo)
!          c2p = sqrt(gap/gbp/velp)

c2m = gam/gbm/velm
c2o = gao/gbo/velo
c2p = gap/gbp/velp

          ! imaginary coefficients
!          aam = opsign*cmplx(0.,sign(dz*0.5/ws*c2o*c1o*c1m*c2m/dh/dh, 1.-vref/velm))
!          aao = opsign*cmplx(0.,sign(dz*0.5/ws*c2o*c1o*c1o*c2o/dh/dh, 1.-vref/velo))
!          aap = opsign*cmplx(0.,sign(dz*0.5/ws*c2o*c1o*c1p*c2p/dh/dh, 1.-vref/velp))

aam = opsign*cmplx(0.,sign(dz*0.5/ws*c1m*c2o/dh/dh, 1.-vref/velm))
aao = opsign*cmplx(0.,sign(dz*0.5/ws*c1o*c2o/dh/dh, 1.-vref/velo))
aap = opsign*cmplx(0.,sign(dz*0.5/ws*c1p*c2o/dh/dh, 1.-vref/velp))

          ! real coefficients
!          bbm = trick + c1o*c1m/dh/dh/ws/ws
!          bbo = trick + c1o*c1o/dh/dh/ws/ws
!          bbp = trick + c1o*c1p/dh/dh/ws/ws

bbm = trick + c1m/dh/dh/ws/ws
bbo = trick + c1o/dh/dh/ws/ws
bbp = trick + c1p/dh/dh/ws/ws

          !sub-diagoal
          ah(ih) = bbm - aam
          !super-diagonal
          ch(ih) = bbp - aap
          !diagonal
          bh(ih) = 1 - (bbo-aao)
          !boundary condition
!          if (ih==1 .or. ih==nh) bh(ih) = 1. - (1.+bndp)*(bbo-aao)
       end do

    else
       do ih=1, nh
          velo = velh(ih)
          ! alph[v] = alph * |1-c/v|
          gao = alph*abs(1.-vref/velo)
          select case (dip)
            case (45)
                 ! beta[v] = beta*(c*c/v/v+c/v+1)
                 gbo = beta*(vref*vref/velo/velo+vref/velo+1.)
            case (65)
                 ! beta[v] = beta*(c*c/v/v+1)
                 gbo = beta*(vref*vref/velo/velo+1.)
            case default
                 write(0,*) "ERROR: dip has to be either 45 or 65 for FFD"
                 stop
          end select
          ! C1 = v sqrt(beta[v])
          c1o = velo*sqrt(gbo)
          ! C2 = alph[v]/beta[v]/v
          c2o = sqrt(gao/gbo/velo)
          ! imaginary coefficients
          aao = opsign*cmplx(0.,sign(dz*0.5/ws*c2o*c1o*c1o*c2o/dh/dh, 1.-vref/velo))
          ! real coefficients
          bbo = trick + c1o*c1o/dh/dh/ws/ws
          ! sub-diagoal
          ah(ih) = bbo - aao
          ! super-diagonal
          ch(ih) = bbo - aao
          ! diagonal
          bh(ih) = 1 - 2.*(bbo-aao)
          ! boundary condition
          if (ih==1 .or. ih==nh) bh(ih) = 1. - (1.+bndp)*(bbo-aao)
       end do
    end if
    ah(1 ) = 0.
    ch(nh) = 0.
    return
  end subroutine
  !!-------------------------------------------------------------------------------------------------------- 
  !!.. bulletproof FFD, for some "unkonwn" reason, 
  !!.. it seems to be even less stable than conventional FFD
  !!.. boundary condition, evanescent waves seem to be playing important roles
  !!--------------------------------------------------------------------------------------------------------
  subroutine ffd_sysform(adj, ah, bh, ch, velh, vref, ws, dh, dz, alph, beta, trick, bndp, dip, nh, opsign, bullet)
    logical                :: adj
    complex, dimension(nh) :: ah, bh, ch
    real   , dimension(nh) :: velh
    real                   :: vref, ws, dh, dz
    real                   :: alph, beta, trick
    complex                :: bndp
    integer                :: dip, nh, opsign
    logical                :: bullet

    integer                :: ih
    real                   :: velo, velm, velp, c1o, c1m, c1p, c3o, c3m, c3p
    complex                :: aah, bbh, aao, bbo, aam, bbm, aap, bbp

    ah = 0; bh = 0; ch = 0

    if (bullet) then
       do ih=1, nh
          velo = velh(ih)
          velm = velh(max(1   ,ih-1))
          velp = velh(min(ih+1,nh  ))
!         coeffcients for C1, 
!         C1 = v sqrt(beta[v]), 
          select case (dip)
            case (45)
!                beta[v] = alph*(c*c/v/v + c/v + 1) for 45 degree
                 c1m = sqrt(beta*(vref*vref+vref*velm+velm*velm))
                 c1o = sqrt(beta*(vref*vref+vref*velo+velo*velo))
                 c1p = sqrt(beta*(vref*vref+vref*velp+velp*velp))
            case (65)
!                beta[v] = beta*(c*c/v/v + 1) for 65 degree
                 c1m = sqrt(beta*(vref*vref+velm*velm))
                 c1o = sqrt(beta*(vref*vref+velo*velo))
                 c1p = sqrt(beta*(vref*vref+velp*velp))
            case (75)
                 c1m = sqrt(beta*(vref*vref+velm*velm))
                 c1o = sqrt(beta*(vref*vref+velo*velo))
                 c1p = sqrt(beta*(vref*vref+velp*velp))
            case default
                 write(0,*) "ERROR: dip has to be either 45 or 65 for FFD"
                 stop
          end select
!         coeffcients for C3,
!         C3 = C2*C1 = sqrt(v*alph[v])
!         alph[v] = alph*abs(1-c/v)
!          c3m = sqrt(velm*alph*abs(1-vref/velm))
!          c3o = sqrt(velo*alph*abs(1-vref/velo))
!          c3p = sqrt(velp*alph*abs(1-vref/velp))
!         imaginary coefficients
!          aam = cmplx(0., opsign*sign(dz*0.5/ws*c3o*c3m/dh/dh,(1-vref/velo)))
!          aao = cmplx(0., opsign*sign(dz*0.5/ws*c3o*c3o/dh/dh,(1-vref/velo)))
!          aap = cmplx(0., opsign*sign(dz*0.5/ws*c3o*c3p/dh/dh,(1-vref/velo)))

          c3m = sqrt(velm)
          c3o = sqrt(velo)
          c3p = sqrt(velp)
          aam = opsign*cmplx(0.,1.)*dz*0.5/ws/dh/dh*alph*(1-vref/velm)*c3o*c3m
          aao = opsign*cmplx(0.,1.)*dz*0.5/ws/dh/dh*alph*(1-vref/velo)*c3o*c3o
          aap = opsign*cmplx(0.,1.)*dz*0.5/ws/dh/dh*alph*(1-vref/velp)*c3o*c3p

!         real coefficients
          bbm = trick + c1o*c1m/dh/dh/ws/ws
          bbo = trick + c1o*c1o/dh/dh/ws/ws
          bbp = trick + c1o*c1p/dh/dh/ws/ws
!         sub-diagoal
          ah(ih) = bbm - aam
!         super-diagonal
          ch(ih) = bbp - aap
!         diagonal
          bh(ih) = 1 - 2.*(bbo-aao)
!         boundary condition
          if (ih==1 .or. ih==nh) bh(ih) = 1. - (1.+bndp)*(bbo-aao)
       end do

    else

       do ih=1, nh
          velo = velh(ih)
!         coeffcients for C1, 
!         C1 = v sqrt(beta[v]), 
          select case (dip)
            case (45)
!                beta[v] = alph*(c*c/v/v + c/v + 1) for 45 degree
                 c1o = sqrt(beta*(vref*vref+vref*velo+velo*velo))
            case (65)
!                beta[v] = beta*(c*c/v/v + 1) for 65 degree
                 c1o = sqrt(beta*(vref*vref+velo*velo))
            case (75)
                 c1o = sqrt(beta*(vref*vref+velo*velo))
            case default
                 write(0,*) "ERROR: dip has to be either 45 or 65 for FFD"
                 stop
          end select
!         coeffcients for C3,
!         C3 = C2*C1 = sqrt(v*alph[v])
!         alph[v] = alph*abs(1-c/v)
!          c3o = sqrt(velo*alph*abs(1-vref/velo))
!         imaginary coefficients
!          aao = cmplx(0., opsign*sign(dz*0.5/ws*c3o*c3o/dh/dh,(1-vref/velo)))
          aao = opsign*cmplx(0.,1.)*dz*0.5/ws/dh/dh*alph*(velo-vref)
!         real coefficients
          bbo = trick + c1o*c1o/dh/dh/ws/ws
!         sub-diagoal
          ah(ih) = bbo - aao
!         super-diagonal
          ch(ih) = bbo - aao
!         diagonal
          bh(ih) = 1. - 2.*(bbo-aao)
!         boundary condition
          if (ih==1 .or. ih==nh) bh(ih) = 1. - (1.+bndp)*(bbo-aao)
       end do
    end if

    if (adj) then
!      transpose and conjugate the matrix
       do ih=1, nh-1
          ah(1)    = ah(ih+1)
          ah(ih+1) = ch(ih)
          ch(ih)   = ah(1)
       end do
       ah = conjg(ah)
       bh = conjg(bh)
       ch = conjg(ch)
    end if

    ah(1 ) = 0.
    ch(nh) = 0.
    return
  end subroutine

  !!-----------------------------------------------------------------------
  !!.. Phase-shift kernel
  !!   Add Q in the phase shift (not in use)
  !!-----------------------------------------------------------------------
!   subroutine phase_shift(adj, wfld_slice, kx2, ky2, vref, qref, ws, dz, nx, ny, opsign)!!!Add qref
!     logical                 :: adj
!     complex, dimension(:,:) :: wfld_slice
!     real   , dimension(:)   :: kx2
!     real   , dimension(:)   :: ky2
!     real                    :: vref, ws
!     real                    :: dz
!     integer                 :: nx, ny
!     integer, intent(in)     :: opsign
! 
!     integer                 :: ix, iy, psign
!     complex                 :: kz, kz2, phase          ! SSR becomes complex
!     real                    :: wov
!     real      		    :: qref              !qref
! 
!     if (adj) then
!        psign = -opsign
!     else 
!        psign =  opsign
!     end if
! 
!     wov = ws / vref
! 
!     do iy=1, ny
!     do ix=1, nx
!        kz2 = (1.+(0.0, 1.0)/2/qref)**2 - kx2(ix)/wov/wov - ky2(iy)/wov/wov    !SSP contains (1+i/2/Q)^2
! !       if (kz2 >= 0.) then   !!!!!!Phase shift without Q
!           kz = sqrt(kz2)     !take the square root of the a complex number
!           phase = psign*wov*kz*dz 
! 	  wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp((0.0, 1.0)*phase)  !  phase shift 
! 
! !!!!!!phase shift without Q
!           !wfld_slice(ix,iy) = wfld_slice(ix,iy)*cmplx(cos(phase),sin(phase))
! !       else !if (kz2 >= -1.) then
! !          kz2   = max(-1., kz2)
! !          kz    = sqrt(-kz2)
! !          phase = - wov*kz*dz
! !          wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp(phase)
! !       end if
!     end do
!     end do
!     return
!   end subroutine

  !!-----------------------------------------------------------------------
  !!.. Phase-shift kernel with Q (Bob)  
  !!-----------------------------------------------------------------------
!   subroutine phase_shift(adj, wfld_slice, kx2, ky2, vref, qref, ws, dz, nx, ny, opsign, qcomp, qcomp_clip,p1, p2, p3 ,p4) !Add qref
!     logical                 :: adj
!     complex, dimension(:,:) :: wfld_slice
!     real   , dimension(:)   :: kx2
!     real   , dimension(:)   :: ky2
!     real                    :: vref, ws
!     real                    :: dz
!     integer                 :: nx, ny
!     integer, intent(in)     :: opsign
!     integer                 :: ix, iy, psign
! 
!     complex                 :: kz, kz2			!!! real -> complex
!     complex                 :: phase, wov		!!! real -> complex
!     real		    :: qref 			!!! Add qref
!     real		    :: wref			!!! Add reference frequency
!     complex		    :: cvref			!!! Add reference velocity with Q
!     real		    :: pi			!!! add pi 
!     complex		    :: cindex			!!! complex number i
!     complex		    :: eva
!     logical	    	    :: qcomp
!     real	    	    :: qcomp_clip
!     real 		    :: p1, p2, p3 , p4
!     complex, allocatable,dimension(:) :: phasev,evav
!     real, allocatable,dimension(:) :: kzv
!     real :: pwov
!     if (adj) then
!        psign = -opsign
!     else 
!        psign =  opsign
!     end if
!     allocate(kzv(nx),evav(nx),phasev(nx))
!     
!     !!!!!!!!wov=ws/vref  -> wov=ws/cvref
! 
!     pi=3.14159265
!     wref=1.*pi   !wref=1./1000.
!     cindex=(0.0,1.0)
!    ! cvref=vref/(1.-1./pi/(10.**qref)*log(ws/wref))/(1.+cindex/2./(10.**qref))
!     cvref=vref/(1.+cindex/2./(10.**qref))
!     wov = ws / cvref
!     pwov=psign*real(wov*kz*kz) 
!     !!!!!!!!end change
! if(qcomp) then
!     do iy=1, ny
!     do ix=1, nx
! !       kzv(ix) = 1. - kx2(ix)/wov/wov - ky2(iy)/wov/wov
! !       kz2 = 1. - kx2(ix)/wov/wov - ky2(iy)/wov/wov
!  !      kz=c_sqrt(kz2)
!         kzv=1
! !       kzv = 1. - kx2(ix)/wov/wov - ky2(iy)/wov/wov
! !       kz=c_sqrt(kz2)
!     end do
! !     end do 
! !       kzv = c_sqrt(kzv)     !take the square root of the a complex number
! !       kzv = kzv     !take the square root of the a complex number
! !       eva=real(wov)*real(wov)-kx2(ix)-ky2(iy)
!        evav=real(wov)*real(wov)-kx2-ky2(iy)
! !       phase=psign*real(wov*kz*dz)-(0.,1.)*(min(qcomp_clip,(aimag(wov*kz*dz)-aimag(c_sqrt(eva)*dz))))
! !       phasev=psign*real(wov*kz*dz)-(0.,1.)*(min(qcomp_clip,(aimag(wov*kz*dz)-aimag(evav*dz))))
!        phasev=pwov-(0.,1.)*aimag(wov*kz*kz)-aimag(evav*dz)
!        wfld_slice(:,iy) = wfld_slice(:,iy)*exp(cindex*phasev)  !  phase shift 
! !       wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp(cindex*phase)  !  phase shift 
!     end do
!   else
! 
!     do iy=1, ny
!     do ix=1, nx
!        kz2 = 1. - kx2(ix)/wov/wov - ky2(iy)/wov/wov
!        kz = c_sqrt(kz2)     !take the square root of the a complex number
! 
! 	eva=real(wov)*real(wov)-kx2(ix)-ky2(iy)
! 
! ! #######attenuation-evanescent
! 		phase=psign*real(wov*kz*dz)+(0.,1.)*aimag(wov*kz*dz )
!       	 	wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp(cindex*phase)  !  phase shift 
!  !	if (real(exp(cindex*phase)).gt.1.8 .and. adj) then
! !		write(0,*)"******************debugx*******************************"
! ! 		write(0,*)"exp(cindex*phase)",exp(cindex*phase)
! !		write(0,*)"evanescent",real(wov)*real(wov)-kx2(ix)-ky2(iy), real(wov)*real(wov)-aimag(wov)*aimag(wov)-kx2(ix)-ky2(iy),real(wov)*real(wov)+aimag(wov)*aimag(wov)-kx2(ix)-ky2(iy)
! !		STOP
! ! 		write(0,*)"*******************************************************"	
! ! 	end if	
! 
! 
!     end do
!     end do
!   end if
! deallocate(evav,phasev,kzv)
!     return
!   end subroutine

 !!-----------------------------------------------------------------------
  !!.. Phase-shift kernel with Q (Yi Shen)  
  !!-----------------------------------------------------------------------
  subroutine phase_shift(adj, wfld_slice, kx2, ky2, vref, qref, ws, dz, nx, ny, opsign, qcomp, qcomp_clip,p1, p2, p3 ,p4) !Add qref
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:)   :: kx2
    real   , dimension(:)   :: ky2
    real                    :: vref, ws
    real                    :: dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign
    integer                 :: ix, iy, psign

    complex                 :: kz, kz2			!!! real -> complex
    complex                 :: phase, wov		!!! real -> complex
    real		    :: qref 			!!! Add qref
    real		    :: wref			!!! Add reference frequency
    complex		    :: cvref			!!! Add reference velocity with Q
    real		    :: pi			!!! add pi 
    complex		    :: cindex			!!! complex number i
    complex		    :: eva
    logical	    	    :: qcomp
    real	    	    :: qcomp_clip
    real 		    :: p1, p2, p3 , p4
    complex		    :: amp, eva_ref, amp_ref    !!! amplitude compensation 
    real		    :: kz_ref, kz2_ref          !!! amplitude compensation
    real		    :: k1, k2, kh2, kh, s       !!! for hamming window

    if (adj) then
       psign = -opsign
    else 
       psign =  opsign
    end if
    
    !!!!!!!!wov=ws/vref  -> wov=ws/cvref

    pi=3.14159265
   ! wref=1000.*pi   !wref=200*pi
    cindex=(0.0,1.0)

    wref=65*pi
    cvref=vref/(1.-1./pi/(10.**qref)*log(ws/wref))/(1.+cindex/2./(10.**qref))
    ! cvref=vref/(1.+cindex/2./(10.**qref))
    wov = ws / cvref

    k1=sin(qcomp_clip*pi/180.0)! acos( limit ang) 
    !k2=1.0 ! 90 degree; evanescent wave
    k2=sin((qcomp_clip+10)*pi/180.0)

    !! calculate the compensated amplitude at the limit angle
    kz2_ref=1. - k1*k1
    kz_ref=sqrt(kz2_ref)
    eva_ref=real(wov)*real(wov)-k1*k1*real(wov)*real(wov)
    amp_ref=aimag(wov*kz_ref*dz)-aimag(c_sqrt(eva_ref)*dz)

    do iy=1, ny
    do ix=1, nx
       kz2 = 1. - kx2(ix)/wov/wov - ky2(iy)/wov/wov
       kz = c_sqrt(kz2)     !take the square root of the a complex number


	eva=real(wov)*real(wov)-kx2(ix)-ky2(iy)
	amp=aimag(wov*kz*dz)-aimag(c_sqrt(eva)*dz)

	!!! param for hamming window
	kh2=kx2(ix)/real(wov)/real(wov)+ky2(iy)/real(wov)/real(wov)
	kh=sqrt(kh2)


! #######attenuation-evanescent
	if(qcomp) then   !!! hamming window of amp
		if( kh.ge. k1 .and. kh .lt. k2 .and. amp .ne. 0.0) then
			s=0.5+0.5*cos(pi*(k1-kh)/(k2-k1))
			amp=amp_ref*s
		else if (kh .ge. k2) then
			amp=0.0;
		end if 
    	!	if(kh .ge. k1) then
	!		amp=0.0
	!	end if
	!	phase=psign*real(wov*kz*dz)-(0.,1.)*amp

		phase=psign*real(wov*kz*dz)-(0.,1.)*amp+(0.,1.)*aimag(c_sqrt(eva)*dz)
			
	!	phase=psign*real(wov*kz*dz)-(0.,1.)*(cmpmin(qcomp_clip,(aimag(wov*kz*dz)-aimag(c_sqrt(eva)*dz))))
	!	phase=psign*real(wov*kz*dz)-(0.,1.)*(softclip(qcomp_clip,(aimag(wov*kz*dz)-aimag(c_sqrt(eva)*dz))))	
!############################
	else
		phase=psign*real(wov*kz*dz)+(0.,1.)*aimag(wov*kz*dz )
	end if
      	 	wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp(cindex*phase)  !  phase shift 
!  	if (real(exp(cindex*phase)).gt.1.8 .and. adj) then
! 		write(0,*)"******************debugx*******************************"
!  		write(0,*)"exp(cindex*phase)",exp(cindex*phase)
! 		write(0,*)"evanescent",real(wov)*real(wov)-kx2(ix)-ky2(iy), real(wov)*real(wov)-aimag(wov)*aimag(wov)-kx2(ix)-ky2(iy),real(wov)*real(wov)+aimag(wov)*aimag(wov)-kx2(ix)-ky2(iy)
! 		STOP
!  		write(0,*)"*******************************************************"	
!  	end if	


    end do
    end do




    return
  end subroutine


  !!-------------------------------------------------------------------------
  !!.. Split-step correction with Q (Yi Shen)
  !!-------------------------------------------------------------------------
  subroutine ssf_correct(adj, wfld_slice, vmod_slice,qmod_slice, vref, qref, ws, dz, nx, ny, opsign)  !!! add qmod_slice, qref
    logical                 :: adj
    complex, dimension(:,:) :: wfld_slice
    real   , dimension(:,:) :: vmod_slice
    real                    :: vref
    real                    :: ws
    real                    :: dz
    integer                 :: nx, ny
    integer, intent(in)     :: opsign
    integer                 :: ix, iy, psign

    real   , dimension(:,:) :: qmod_slice 		!!! Add qmod_slice
    complex                 :: split, wov		!!! real -> complex
    real		    :: qref 			!!! Add qref
    complex		    :: cvref			!!! Add reference velocity with Q
    complex,  allocatable   :: cvmod_slice(:,:)		!!! Add velocity with Q

    real		    :: wref			!!! Add reference frequency
    real		    :: pi			!!! add pi 
    complex		    :: cindex			!!! complex number i


    if (adj) then
       psign = -opsign
    else
       psign =  opsign
    end if

    !!!!!!!!wov=ws/vref  -> wov=ws/cvref


    pi=3.14159265
    wref=65*pi  !wref=1./1000.
    cindex=(0.0,1.0)
    cvref=vref/(1.-1./pi/(10.**qref)*log(ws/wref))/(1.+cindex/2./(10.**qref))
    !cvref=vref/(1.+cindex/2./10.**qref)
    wov = ws / cvref
    allocate(cvmod_slice(nx,ny))   !!! allocate

    do iy=1, ny
    do ix=1, nx
    !!!!!!!!vmod->cvmod
       cvmod_slice(ix,iy)=vmod_slice(ix,iy)/(1.-1./pi/(10.**qmod_slice(ix,iy))*log(ws/wref))/(1.+cindex/2./(10.**qmod_slice(ix,iy)))
!	cvmod_slice(ix,iy)=vmod_slice(ix,iy)/(1.+cindex/2./(10.**qmod_slice(ix,iy)))
       split = psign*wov*dz*(cvref/cvmod_slice(ix,iy)-1.)
       wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp(cindex*split)
!		if(abs(sum(wfld_slice)) .gt. 10000.) then
!		write(0,*)"correct",split

!		STOP
!		end if
! 		if(real(exp(cindex*split)) .gt. 1.) then
! 		write(0,*)"else if",cvref, wov, split,psign, exp(cindex*split)
! 
! 		STOP
! 		end if
    end do 
    end do

    deallocate(cvmod_slice)    !!! Deallocate
    return
  end subroutine

  !!-------------------------------------------------------------------------
  !!.. Split-step vel correction  (Yi Shen)  --- not used
  !!-------------------------------------------------------------------------
!   subroutine ssf_vel_correct(adj, wfld_slice, vmod_slice, qmod_slice, vref,qref,wref ws, dz, nx, ny, opsign) ! and qmod_slice, qref,wref
!     logical                 :: adj
!     complex, dimension(:,:) :: wfld_slice
!     real   , dimension(:,:) :: vmod_slice
!     real                    :: vref
!     real                    :: ws
!     real                    :: dz
!     integer                 :: nx, ny
!     integer, intent(in)     :: opsign
!  
!     integer                 :: ix, iy, psign
!   !  real                    :: split, wov
! 
! !!!!!!Add q
!     complex, dimension(:,:) :: vtilde_slice  ! velocity with q
!     complex		    :: vtilde_ref    ! vrefer with q
!     real 		    :: pi           ! add in vel
!     real  		    :: wref
!     real		    :: ws
!     complex		    :: split
! 
!     pi=2*acos(0)
! 
!     if (adj) then
!        psign = -opsign
!     else
!        psign =  opsign
!     end if
! 
!     vtilde_ref=vref/(1-1/pi/qref*log(ws/wref))/(1+(0.,1.)/2/qref)
!    ! wov = ws / vref
!     
!     do iy=1, ny
!     do ix=1, nx
! 	vtilde_slice(ix,iy)=vmod_slice/(1-1/pi/qmod_slice(ix,iy)*log(ws/wref))/(1+(0.,1.)/2/qmod_slice(ix,iy))
!        ! split = psign*wov*dz*(vref/vmod_slice(ix,iy)-1.)
!   	split = psign*ws/vtilde_ref*dz*(vtilde_ref/vtilde_slice(ix,iy)-1.)
!         wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp((0.,1.)*split)
!     end do 
!     end do
!     return
!   end subroutine

  !!-------------------------------------------------------------------------
  !!.. Split-step q correction (Yi Shen) --- not used
  !!-------------------------------------------------------------------------
!   subroutine    ssf_q_correct(adj, wfld_slice, vmod_slice,qmod_slice, vref,qref, ws, dz, nx, ny, opsign)
!     logical                 :: adj
!     complex, dimension(:,:) :: wfld_slice
!     real   , dimension(:,:) :: vmod_slice,qmod_slice
!     real                    :: vref,qref
!     real                    :: ws
!     real                    :: dz
!     integer                 :: nx, ny
!     integer, intent(in)     :: opsign
!  
!     integer                 :: ix, iy, psign
!     real                    :: wov
!     complex	 	    :: split
! 
!     if (adj) then
!        psign = -opsign
!     else
!        psign =  opsign
!     end if
!     wov = ws / vref
!     
!     do iy=1, ny
!     do ix=1, nx
!        split = psign*wov*dz*(0.,1.)/2*(vref/vmod_slice(ix,iy)/qmod_slice(ix,iy)-1./qref)
!       ! wfld_slice(ix,iy) = wfld_slice(ix,iy)*cmplx(cos(split),sin(split))
! 	wfld_slice(ix,iy) = wfld_slice(ix,iy)*exp((0.,1.)*split)
!     end do 
!     end do
!     return
!   end subroutine



  !!--------------------------------------------------------------------------
  !!.. Fourier finite-difference correction
  !!--------------------------------------------------------------------------
  subroutine ffd_correct(adj, wfld_slice, vmod_slice, alph, beta, vref, ws, dx, dy, dz, trick, bndp, &
                         nx, ny, nstep, dip, opsign, bullet, rwrk1d, cwrk1d)
    logical                          :: adj
    complex, dimension(:,:)          :: wfld_slice
    real   , dimension(:,:)          :: vmod_slice
    real                             :: vref, ws
    real                             :: dx, dy, dz, trick
    complex                          :: bndp
    integer                          :: nx, ny, dip
    integer, intent(in)              :: opsign
    logical                          :: bullet
    real   , dimension(nx+ny)        :: rwrk1d
    complex, dimension(nx*5+ny*5)    :: cwrk1d

    real                             :: alph(5), beta(5)
    integer                          :: ix, iy, istep, nstep 
    integer                          :: lrx1, lry1
    integer                          :: lcx1, lcx2, lcx3, lcx4, lcx5
    integer                          :: lcy1, lcy2, lcy3, lcy4, lcy5

    !!.. asign work space
    lcx1 = 1            !!.. wfdx
    lcx2 = lcx1 + nx    !!.. ahx
    lcx3 = lcx2 + nx    !!.. bhx
    lcx4 = lcx3 + nx    !!.. chx
    lcx5 = lcx4 + nx    !!.. rhx
    lcy1 = lcx5 + nx    !!.. wfdy
    lcy2 = lcy1 + ny    !!.. ahy
    lcy3 = lcy2 + ny    !!.. bhy
    lcy4 = lcy3 + ny    !!.. chy
    lcy5 = lcy4 + ny    !!.. rhy

    lrx1 = 1
    lry1 = lrx1 + nx

    if (adj) then
!      solve FFD correction in y direction
!      -----------------------------------
       if (ny > 1) then
          do ix=1, nx
             cwrk1d(lcy1:lcy2-1) = wfld_slice(ix,:)
             rwrk1d(lry1:nx+ny ) = vmod_slice(ix,:)
             call ffd_sysform(adj, cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), rwrk1d(lry1), vref, ws, dy, dz, &
                              alph(1), beta(1), trick, bndp, dip, ny, opsign, bullet)
             call ifd_dfrterm(adj, cwrk1d(lcy1), cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), cwrk1d(lcy5), ny)
             wfld_slice(ix,:) = cwrk1d(lcy1:lcy2-1)
          end do
       end if
!      solve FFD correction in x direction
!      -----------------------------------
       if (nx > 1) then
          do iy=1, ny
             cwrk1d(lcx1:lcx2-1) = wfld_slice(:,iy)
             rwrk1d(lrx1:lry1-1) = vmod_slice(:,iy)
             call ffd_sysform(adj, cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), rwrk1d(lrx1), vref, ws, dx, dz, &
                              alph(1), beta(1), trick, bndp, dip, nx, opsign, bullet)
             call ifd_dfrterm(adj, cwrk1d(lcx1), cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), cwrk1d(lcx5), nx)
             wfld_slice(:,iy) = cwrk1d(lcx1:lcx2-1)
          end do
       end if

    else 
!      solve FFD correction in x direction
!      -----------------------------------
       if (nx > 1) then
          do iy=1, ny
             cwrk1d(lcx1:lcx2-1) = wfld_slice(:,iy)
             rwrk1d(lrx1:lry1-1) = vmod_slice(:,iy)
             call ffd_sysform(adj, cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), rwrk1d(lrx1), vref, ws, dx, dz, &
                              alph(1), beta(1), trick, bndp, dip, nx, opsign, bullet)
             call ifd_dfrterm(adj, cwrk1d(lcx1), cwrk1d(lcx2), cwrk1d(lcx3), cwrk1d(lcx4), cwrk1d(lcx5), nx)
             wfld_slice(:,iy) = cwrk1d(lcx1:lcx2-1)
          end do
       end if
!      solve FFD correction in y direction
!      -----------------------------------
       if (ny > 1) then
          do ix=1, nx
             cwrk1d(lcy1:lcy2-1) = wfld_slice(ix,:)
             rwrk1d(lry1:nx+ny ) = vmod_slice(ix,:)
             call ffd_sysform(adj, cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), rwrk1d(lry1), vref, ws, dy, dz, &
                              alph(1), beta(1), trick, bndp, dip, ny, opsign, bullet)
             call ifd_dfrterm(adj, cwrk1d(lcy1), cwrk1d(lcy2), cwrk1d(lcy3), cwrk1d(lcy4), cwrk1d(lcy5), ny)
             wfld_slice(ix,:) = cwrk1d(lcy1:lcy2-1)
          end do
       end if
    end if
    return
  end subroutine

  !!---------------------------------------------------------------------------
  !!.. complex tridiagonal solver
  !!---------------------------------------------------------------------------
  !!.. a(n), sub-diagonal elemnets, only elements from 2:n are used
  !!.. b(n), diagonal elemetns
  !!.. c(n), super-diagonal elements, only elements from 1:n-1 are used
  !!
  !!   /  b(1), c(1)                                     \ / u(  1) \     / r(  1) \ 
  !!   |  a(2), b(2), c(2)                               | | u(  2) |     | r(  2) |
  !!   |        a(3), b(3), c(3)                         | | u(  3) |  =  | r(  3) |
  !!   |              ... ...                            | |  ...   |     |  ...   |
  !!   |                          a(n-1), b(n-1), c(n-1) | | u(n-1) |     | r(n-1) |
  !!   \                                  a(  n), b(  n) / \ u(  n) /     \ r(  n) /
! ************************
! NOTE: r will be modified
! ************************
  subroutine ctridiag(a,b,c,r,u,n)
    integer :: n
    complex :: a(n), b(n), c(n), r(n), u(n)
    integer :: j
    complex :: bet

    bet  = b(1)
    u(1) = r(1)/bet
    do j=2, n
       r(j-1) = c(j-1)/bet
       bet    = b(j)-a(j)*r(j-1)
       u(j) = (r(j)-a(j)*u(j-1))/bet
    end do
    do j=n-1, 1, -1
       u(j) = u(j) - r(j)*u(j+1)
    end do
  end subroutine

! ************************
! nothing will be modified
! ************************
  subroutine ctridiag2(a,b,c,r,u,n)
    integer            :: n
    complex            :: a(n), b(n), c(n), r(n), u(n)
    integer            :: j
    integer, parameter :: NMAX = 5000
    complex            :: bet, gam(NMAX)
    !if (cabs(b(1)) == 0.) stop "ctirdag: rewrite equations"
    bet  = b(1)
    u(1) = r(1)/bet
    do j=2, n
       gam(j) = c(j-1)/bet
       bet    = b(j)-a(j)*gam(j)
       !if (cabs(bet) == 0.) stop "ctridag failed"
       u(j) = (r(j)-a(j)*u(j-1))/bet
    end do 
    do j=n-1,1,-1
       u(j) = u(j) - gam(j+1)*u(j+1)
    end do
    return
  end subroutine

! ************************
! b, c, u will be modified
! ************************
  subroutine ctridiag3(a,b,c,u,n)
    integer :: n
    complex :: a(n), b(n), c(n), u(n)
    integer :: i
    c(1) = c(1)/b(1)
    u(1) = u(1)/b(1)
    do i=2,n
       b(i) = b(i) - a(i)*c(i-1)
       if (i<n) then
          c(i) = c(i)/b(i)
       end if
       u(i) = (u(i)-a(i)*u(i-1))/b(i)
    end do
    do i=n-1,1,-1
       u(i) = u(i) - c(i)*u(i+1)
    end do
    return
  end subroutine

  !!-----------------------------------------------------------------------------------------
  !!.. set coefficients for implicit finite-difference extrapolator
  !!.. based on Zaitian Ma's higher order splitting method
  !!.. the coefficients are optimized based on Lee and Suh's paper
  !!.. Ma, Z., 1981, Finite-difference migration with higher-order approximation:
  !!.. Presented at the 1981 joint meeting of the China Geophysical Society and
  !!.. Soc. of Expl. Geophys., Beijing, China
  !!.. Lee, M. W. and S. Y., Suh, 1985, Optimization of one-way wave equations,
  !!.. Geophysics, 50, 1634-1637.
  !!-----------------------------------------------------------------------------------------
  subroutine ifd_coeffts(alph, beta, nstep, dip)
    real    :: alph(5), beta(5)
    integer :: nstep
    integer :: dip
    select case (dip)
      case (15)
           alph(1) = 0.500000000; beta(1) = 0.000000000
           nstep = 1

      case (45)
           alph(1) = 0.500000000; beta(1) = 0.250000000
           nstep = 1

      case (65)
           alph(1) = 0.478242060; beta(1) = 0.376369527
           nstep = 1

      case (79)
           alph(1) = 0.457500000; beta(1) = 0.457500000
           nstep = 1

      case (80)
           alph(1) = 0.457289566; beta(1) = 0.222691983
           alph(2) = 0.040315157; beta(2) = 0.873981642
           nstep = 2

      case (87)
           alph(1) = 0.414236605; beta(1) = 0.150843924
           alph(2) = 0.081312882; beta(2) = 0.744418059
           alph(3) = 0.004210420; beta(3) = 0.972926132
           nstep = 3

      case (89)
           alph(1) = 0.367013245; beta(1) = 0.105756624
           alph(2) = 0.117592008; beta(2) = 0.614520676
           alph(3) = 0.014853510; beta(3) = 0.919432661
           alph(4) = 0.000523275; beta(4) = 0.994065088
           nstep = 4

      case (90)
           alph(1) = 0.318013812; beta(1) = 0.073588213
           alph(2) = 0.143798076; beta(2) = 0.483340757
           alph(3) = 0.033860918; beta(3) = 0.824918565
           alph(4) = 0.004172967; beta(4) = 0.964827992
           alph(5) = 0.000153427; beta(5) = 0.997370236
           nstep = 5

      case default
           write(0,*)"ERROR: select dip in (15,45,65,79,80,87,89,90) in implicit finite difference"
           STOP

    end select
    return
  end subroutine

  subroutine ffd_coeffts(alph, beta, mmax, dip)
    real    :: alph(5), beta(5)
    real    :: mmax
    integer :: dip 
    integer :: jm
    select case (dip)
!     conventional FFD coefficients
!     -----------------------------
      case (45)
           alph(1) = 0.500000000; beta(1) = 0.250000000

!     FFD based on Pade expansion, using 65 degree coefficients
!     ---------------------------------------------------------
      case (65)
           alph(1) = 0.478242060; beta(1) = 0.376369527

!     FFD based on Pade expansion, globally optimized coeffcients
!     Huang, L. and M., Fehler, 2000, Globally optimized Fourier
!     finite-difference migration method, SEG 2000 Expanded Abstract
!     --------------------------------------------------------------
      case (75)

           jm = nint(1.+(mmax-1.5)/0.5)
           jm = max(1, jm)
           jm = min(jm, 5)
           select case (jm)
             case (1) 
                  alph(1) = 0.488790000; beta(1) = 0.431510000
             case (2)
                  alph(1) = 0.486860000; beta(1) = 0.432970000
             case (3)
                  alph(1) = 0.482300000; beta(1) = 0.436010000
             case (4)
                  alph(1) = 0.478260000; beta(1) = 0.438480000
             case (5)
                  alph(1) = 0.474940000; beta(1) = 0.440320000   
           end select 

      case default
           write(0,*)"ERROR: select dip in (45,65,75) in Fourier finite difference"
    end select
    return
  end subroutine

  integer function rsign(x)
    real :: x
    real :: threshold
    threshold = abs(epsilon(x))
    if (abs(x) <= threshold) then
       rsign = 0
    else if (x > +threshold) then
       rsign = +1
    else if (x < -threshold) then
       rsign = -1
    end if
  end function

  complex function c_sqrt(x)
    complex :: x
    real    :: a,b
    real    :: c,d
    real    :: rp, ip
    real    :: sgn
    a=real(x)
    b=aimag(x)

    c=a+sqrt(a*a+b*b)
    d=-a+sqrt(a*a+b*b)

    if (b >= 0.) then
	sgn=1.
    else
	sgn=-1.
    end if

    rp=sqrt(c/2)
    ip=sgn*sqrt(d/2)

    c_sqrt=rp*(1.,0.)+(0.,1.)*ip

  end function


  real function cmpmin(x,y)
    real :: x,y

    if (x >= y) then
	cmpmin=y
    else
	cmpmin=x
    end if
  end function

  real function softclip(hardmax,r)
       real :: softmax, r
	real:: hardmax,fac
	fac=10.
	softmax=hardmax/fac
	if (hardmax>r) then	
		softclip=r
	else 
		softclip=softmax*r/softmax/sqrt(1.+r*r/softmax/softmax)*sqrt(1.+fac*fac)
	end if
  end function




end module
