!!-----------------------------------------------------------------
!! Author: Yaxun Tang, tang@sep.stanford.edu
!!-----------------------------------------------------------------
module bwi_glocal_mod
  use  bwi_struct_mod
  use  base_progrd_mod
  use  base_outgrd_mod
  use  base_velgrd_mod
  use  base_qgrd_mod
  implicit none

contains

  subroutine bwi_glocal(pgrid, ogrid, vgrid, qgrid, dgrid, igrid, hgrid, ish)   !!! Add qgrid
    type(pro_grid) :: pgrid
    type(out_grid) :: ogrid
    type(vel_grid) :: vgrid
    type(q_grid) :: qgrid	!!!qgrid
    type(dat_grid) :: dgrid
    type(img_grid) :: igrid
    type(hes_grid) :: hgrid
    integer        :: ish

    integer        :: ishx, ishy
    integer        :: ogrid_extr_step, hgrid_extr_step
    real           :: ogrid_hz_max,    hgrid_z_offd_max

    if (dgrid%hdrsou) then
       dgrid%sx = dgrid%hsou(1,ish)
       dgrid%sy = dgrid%hsou(2,ish)
    else
       ishx = mod((ish-1)/1          , dgrid%asx%n)+1
       ishy = mod((ish-1)/dgrid%asx%n, dgrid%asy%n)+1
       dgrid%sx = dgrid%asx%o + (ishx-1)*dgrid%asx%d
       dgrid%sy = dgrid%asy%o + (ishy-1)*dgrid%asy%d
    end if

    if (dgrid%offset) then
       dgrid%agx%o = dgrid%gxmin_save + dgrid%sx
       dgrid%agy%o = dgrid%gymin_save + dgrid%sy
    end if
    call base_progrd_gen3d(pgrid, dgrid, vgrid)
    call base_outgrd_gen3d(ogrid, igrid, pgrid)
    call base_velgrd_gen3d(vgrid, pgrid, igrid)
    call base_qgrd_gen3d(qgrid, pgrid, igrid)
!    this is for hz gather or Hessian
!    --------------------------------
!    ogrid_hz_max     = ogrid%ahz%o + (ogrid%ahz%n-1)*ogrid%ahz%d
!    ogrid_extr_step  = 1.5 + (ogrid_hz_max-0.)/ogrid%ahz%d
!    hgrid_z_offd_max = hgrid%az_offd%o + (hgrid%az_offd%n-1)*hgrid%az_offd%d
!    hgrid_extr_step  = 1.5 + (hgrid_z_offd_max-0.)/hgrid%az_offd%d

!    pgrid%nzstep = pgrid%nzstep + max(ogrid_extr_step, hgrid_extr_step)
!    pgrid%nzstep = min(pgrid%az%n, pgrid%nzstep)

    return
  end subroutine

  subroutine bwi_wfs_init(adj, add, wfds, csou, dgrid, pgrid)
    logical                   :: adj
    logical                   :: add
    complex, dimension(:,:,:) :: wfds !! (x,y,w)
    complex, dimension(:,:)   :: csou
    type(dat_grid)            :: dgrid 
    type(pro_grid)            :: pgrid

    integer                   :: ix, iy, iws, itr
    integer                   :: indx, indy
    real                      :: gx, gy
    real                      :: wtx, wty, norm
    real                      :: PI

    PI = 3.14159265359

    if (.not.add) then
       if (adj) then
          wfds = 0.
       else
          csou = 0.
       end if
    end if

    if (dgrid%comsou) then
       if (adj) then
          do iy=1, dgrid%agy%n; gy=dgrid%agy%o+(iy-1)*dgrid%agy%d; indy=1.5+(gy-pgrid%ay%o)/pgrid%ay%d
          do ix=1, dgrid%agx%n; gx=dgrid%agx%o+(ix-1)*dgrid%agx%d; indx=1.5+(gx-pgrid%ax%o)/pgrid%ax%d
             itr  = ix + (iy-1)*dgrid%agx%n
             wfds(indx,indy,:) = wfds(indx,indy,:) + conjg(csou(itr,:))
          end do
          end do 
       else
          do iy=1, dgrid%agy%n; gy=dgrid%agy%o+(iy-1)*dgrid%agy%d; indy=1.5+(gy-pgrid%ay%o)/pgrid%ay%d
          do ix=1, dgrid%agx%n; gx=dgrid%agx%o+(ix-1)*dgrid%agx%d; indx=1.5+(gx-pgrid%ax%o)/pgrid%ax%d
             itr  = ix + (iy-1)*dgrid%agx%n
             csou(itr,:) = csou(itr,:) + conjg(wfds(indx,indy,:))
          end do
          end do
       end if
    else 
       !!.. potential error, should use isotropic smoothing
       if (adj) then
          if (dgrid%smtsou) then
             norm = 0.
             do iy=max(pgrid%index_datpro_sy-2,1), min(pgrid%index_datpro_sy+2,pgrid%ay%n); wty=0.5+0.5*cos(PI*real(iy-pgrid%index_datpro_sy)/3)
             do ix=max(pgrid%index_datpro_sx-2,1), min(pgrid%index_datpro_sx+2,pgrid%ax%n); wtx=0.5+0.5*cos(PI*real(ix-pgrid%index_datpro_sx)/3)
                norm = norm + wtx*wty
                wfds(ix,iy,:) = wfds(ix,iy,:) + wtx*wty*conjg(csou(1,:))
             end do
             end do
             wfds = wfds/norm
          else
             wfds(pgrid%index_datpro_sx,pgrid%index_datpro_sy,:) = conjg(csou(1,:))
          end if
       else
          if (dgrid%smtsou) then
             norm = 0.
             do iy=max(pgrid%index_datpro_sy-2,1), min(pgrid%index_datpro_sy+2,pgrid%ay%n); wty=0.5+0.5*cos(PI*real(iy-pgrid%index_datpro_sy)/3)
             do ix=max(pgrid%index_datpro_sx-2,1), min(pgrid%index_datpro_sx+2,pgrid%ax%n); wtx=0.5+0.5*cos(PI*real(ix-pgrid%index_datpro_sx)/3)
                norm = norm + wtx*wty
                csou(1,:) = csou(1,:) + wtx*wty*conjg(wfds(ix,iy,:))
             end do
             end do
             csou(1,:) = csou(1,:)/norm
          else
             csou(1,:) = wfds(pgrid%index_datpro_sx,pgrid%index_datpro_sy,:)
          end if
       end if

    end if
    return
  end subroutine

  subroutine bwi_wfr_init(adj, add, wfdr, crec, dgrid, pgrid)
    logical                   :: adj 
    logical                   :: add
    complex, dimension(:,:,:) :: wfdr
    complex, dimension(:,:)   :: crec
    type(dat_grid)            :: dgrid
    type(pro_grid)            :: pgrid

    integer                   :: ix, iy, itr
    integer                   :: indx, indy 
    real                      :: gx, gy

    if (.not.add) then
       if (adj) then
          wfdr = 0.
       else
          crec = 0.
       end if 
    end if

    if (adj) then
       do iy=1, dgrid%agy%n; gy=dgrid%agy%o+(iy-1)*dgrid%agy%d; indy=1.5+(gy-pgrid%ay%o)/pgrid%ay%d
       do ix=1, dgrid%agx%n; gx=dgrid%agx%o+(ix-1)*dgrid%agx%d; indx=1.5+(gx-pgrid%ax%o)/pgrid%ax%d
           itr  = ix + (iy-1)*dgrid%agx%n
           wfdr(indx,indy,:) = wfdr(indx,indy,:) + crec(itr, :)
       end do
       end do
    else
       do iy=1, dgrid%agy%n; gy=dgrid%agy%o+(iy-1)*dgrid%agy%d; indy=1.5+(gy-pgrid%ay%o)/pgrid%ay%d
       do ix=1, dgrid%agx%n; gx=dgrid%agx%o+(ix-1)*dgrid%agx%d; indx=1.5+(gx-pgrid%ax%o)/pgrid%ax%d
           itr  = ix + (iy-1)*dgrid%agx%n
           crec(itr,:) = crec(itr,:) + wfdr(indx,indy,:)
       end do
       end do
    end if
    return
  end subroutine
end module

