module bwi_grdprc_mod
  use base_struct_mod
  use base_spline_mod
  use base_hlxtyp_mod
  use bwi_fileio_mod
  use bwi_dipcal_mod
  use bwi_poldiv_mod
  implicit none

contains

! scale the gradient by depth or sqrt(depth)
! -------------------------------------------
  integer function bwi_gscale_oper(grdo_, qgrid) result(stat)
    character(len=*)                    :: grdo_
    type(q_grid)                        :: qgrid

    integer                             :: izs
    real                                :: fctr, zs, zmax
    real, dimension(:,:,:), allocatable :: grdo
    double precision                    :: norm

    allocate(grdo(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))
!   read the gradient
!   -----------------
    call bwi_fio_qmod_updt(grdo, qgrid, grdo_, "reed")
!   apply the scaling 
!   -----------------
    zmax = qgrid%az%n*qgrid%az%d
    do izs=1, qgrid%az%n
       zs   = izs*qgrid%az%d
       fctr = sqrt(zs)/sqrt(zmax)
       grdo(:,:,izs) = grdo(:,:,izs)*fctr
    end do

!   normalize the gradient (NEW !!!!!!!!!!!)
!   ----------------------------------------
!    norm = sqrt(sum(dble(grdo)*dble(grdo)))
!    grdo = grdo/(norm+epsilon(norm))

!   write out the scaled gradient
!   -----------------------------
    call bwi_fio_qmod_updt(grdo, qgrid, grdo_, "rite")
    deallocate(grdo)
    stat = 0
  end function

!!!???????????????????????????????????????????????????
!!!???????????????????????????????????????????????????
!!!???????????????????????????????????????????????????
! the size of image and the size of velocity is different
! make sure the share the same grid for dip regularization
  integer function bwi_update_dips(bimg_, dipx, dipy, selx, sely, igrid, vgrid, adip, rectx, recty, rectz, nsmt) result(stat)
    character(len=*)                    :: bimg_
    real   , dimension(:,:,:)           :: dipx
    real   , dimension(:,:,:)           :: dipy
    integer, dimension(:,:,:)           :: selx
    integer, dimension(:,:,:)           :: sely
    type(img_grid)                      :: igrid
    type(vel_grid)                      :: vgrid
    type(axis)                          :: adip
    integer                             :: rectx, recty, rectz, nsmt

    integer                             :: jxb, jxe
    integer                             :: jyb, jye
    integer                             :: jzb, jze
    integer                             :: nrx, nry, nrz, nst, jx, jy 
    integer                             :: iselx, isely, ix, iy, iz
    real, dimension(:,:,:), allocatable :: istk

    allocate(istk(igrid%ax%n,igrid%ay%n,igrid%az%n))

    if (vgrid%ax%d /= igrid%ax%d) call seperr("ERROR: update_dips: vgrid%ax%d /= igrid%ax%d")
    if (vgrid%ay%d /= igrid%ay%d) call seperr("ERROR: update_dips: vgrid%ay%d /= igrid%ay%d")
    if (vgrid%az%d /= igrid%az%d) call seperr("ERROR: update_dips: vgrid%az%d /= igrid%az%d")

!   image indexes
!   -------------
    jxb = vgrid%index_imgvel_xmin; jxe = vgrid%index_imgvel_xmax
    jyb = vgrid%index_imgvel_ymin; jye = vgrid%index_imgvel_ymax
    jzb = vgrid%index_imgvel_zmin; jze = vgrid%index_imgvel_zmax

!   smoothing windows
    nrx = max(1, rectx/2)
    nry = max(1, recty/2)
    nrz = max(1, rectz/2)
    nrx = min(nrx, vgrid%ax%n)
    nry = min(nry, vgrid%ay%n)
    nrz = min(nrz, vgrid%az%n)
    nst = max(1, nsmt/2)

!   read the zero-offset image
!   --------------------------
    call bwi_fio_istk_updt(istk, igrid, bimg_, "reed")

!   estimate the dip field in x
!   ---------------------------
    if (vgrid%ax%n>1) then
       stat = bwi_dipcal_init(igrid%ax,igrid%az,rectx,rectz,nsmt)
       do iy=1, igrid%ay%n
          jy = jyb+iy-1
          stat = bwi_dipcal_oper(istk(:,iy,:), dipx(jxb:jxe,jy,jzb:jze))
       end do
       stat = bwi_dipcal_clos()
       stat = base_tris3d_oper(dipx, nrx, nry, nrz, nst)
       selx = adip%n
       do iz=jzb, jze
       do iy=jyb, jye
       do ix=jxb, jxe
          iselx = 1.5 + (dipx(ix,iy,iz)-adip%o)/adip%d
          if (iselx<1 .or. iselx>adip%n-1) cycle
          selx(ix,iy,iz) = iselx
       end do
       end do
       end do
    end if

!   estimate the dip field in y
!   ---------------------------
    if (vgrid%ay%n>1) then
       stat = bwi_dipcal_init(igrid%ay,igrid%az,recty,rectz,nsmt)
       do ix=1, igrid%ax%n
          jx = jxb+ix-1
          stat = bwi_dipcal_oper(istk(ix,:,:), dipy(jx,jyb:jye,jzb:jze))
       end do
       stat = bwi_dipcal_clos()
       stat = base_tris3d_oper(dipy, nrx, nry, nrz, nst)
       sely = adip%n
       do iz=jzb, jze
       do iy=jyb, jye
       do ix=jxb, jxe
          isely = 1.5 + (dipy(ix,iy,iz)-adip%o)/adip%d
          if (isely<1 .or. isely>adip%n-1) cycle
          sely(ix,iy,iz) = isely
       end do
       end do
       end do
    end if

    deallocate(istk)
    stat = 0
  end function

  integer function bwi_grdmsk_oper(grdo_, gmsk_, qgrid) result(stat)
    character(len=*) :: grdo_, gmsk_
    type(q_grid)   :: qgrid

    real, dimension(:,:,:), allocatable :: grdo, gmsk

    allocate(grdo(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))
    allocate(gmsk(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))
    call bwi_fio_qmod_updt(grdo, qgrid, grdo_, "reed")
    call bwi_fio_qmod_updt(gmsk, qgrid, gmsk_, "reed")
    grdo = grdo*gmsk
	write(0,*), "gmask", maxval(gmsk),minval(gmsk)
    call bwi_fio_qmod_updt(grdo, qgrid, grdo_, "rite")
    deallocate(grdo)
    deallocate(gmsk)  
    stat = 0
  end function

  integer function bwi_dipflt_oper(grdo_, dfcx, dfcy, dfcx0, dfcy0, selx, sely, vgrid) result(stat)
    character(len=*)                    :: grdo_
    type(helix), dimension(:)           :: dfcx
    type(helix), dimension(:)           :: dfcy
    real       , dimension(:)           :: dfcx0
    real       , dimension(:)           :: dfcy0 
    integer    , dimension(:,:,:)       :: selx
    integer    , dimension(:,:,:)       :: sely  
    type(vel_grid)                      :: vgrid
 
    real, dimension(:,:,:), allocatable :: grdo
    real, dimension(:,:)  , allocatable :: tmpx
    real, dimension(:,:)  , allocatable :: tmpy
    integer                             :: ix, iy

    allocate(grdo(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
    allocate(tmpx(vgrid%ax%n,vgrid%az%n)) 
    allocate(tmpy(vgrid%ay%n,vgrid%az%n)) 

    call bwi_fio_vmod_updt(grdo, vgrid, grdo_, "reed")

!   Ax^{-1}Ay^{-1}Ay^{-T}Ax^{-T}
    if (vgrid%ax%n>1) then
       stat = bwi_poldiv_init(dfcx, dfcx0, vgrid%ax%n, vgrid%az%n)
       do iy=1, vgrid%ay%n
          stat = bwi_poldiv_oper(.true. ,.false.,tmpx,grdo(:,iy,:),selx(:,iy,:))
          grdo(:,iy,:) = tmpx
       end do
       stat = bwi_poldiv_clos()
    end if

    if (vgrid%ay%n>1) then
       stat = bwi_poldiv_init(dfcy, dfcy0, vgrid%ay%n, vgrid%az%n)
       do ix=1, vgrid%ax%n
          stat = bwi_poldiv_oper(.true. ,.false.,tmpy,grdo(ix,:,:),sely(ix,:,:))
          stat = bwi_poldiv_oper(.false.,.false.,tmpy,grdo(ix,:,:),sely(ix,:,:))
       end do
       stat = bwi_poldiv_clos()
    end if

    if (vgrid%ax%n>1) then
       stat = bwi_poldiv_init(dfcx, dfcx0, vgrid%ax%n, vgrid%az%n)
       do iy=1, vgrid%ay%n
          tmpx = grdo(:,iy,:) 
          stat = bwi_poldiv_oper(.false.,.false.,tmpx,grdo(:,iy,:),selx(:,iy,:))
       end do
       stat = bwi_poldiv_clos()
    end if 

    call bwi_fio_vmod_updt(grdo, vgrid, grdo_, "rite")
    deallocate(grdo)
    deallocate(tmpx)
    deallocate(tmpy)
    stat = 0
  end function

  integer function bwi_triflt_oper(grdo_, qgrid, rectx, recty, rectz, nsmt) result(stat)
    character(len=*)                    :: grdo_
    type(q_grid)                      :: qgrid
    integer                             :: rectx, recty, rectz, nsmt

    real, dimension(:,:,:), allocatable :: grdo

    allocate(grdo(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))
    call bwi_fio_qmod_updt(grdo, qgrid, grdo_, "reed")
    stat = base_tris3d_oper(grdo, rectx, recty, rectz, nsmt)
    call bwi_fio_qmod_updt(grdo, qgrid, grdo_, "rite")
    deallocate(grdo)
    stat = 0
  end function

  integer function bwi_spline_oper(grdo_, vgrid, axk, ayk, azk) result(stat)
    character(len=*)                    :: grdo_
    type(vel_grid)                      :: vgrid
    type(axis)                          :: axk, ayk, azk

    real, dimension(:,:,:), allocatable :: grdo
    real, dimension(:,:,:), allocatable :: vspl

    allocate(grdo(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
    allocate(vspl(axk%n,ayk%n,azk%n))
    stat = base_spln3d_init(vgrid%ax, vgrid%ay, vgrid%az, axk, ayk, azk)
    call bwi_fio_vmod_updt(grdo, vgrid, grdo_, "reed")
    stat = base_spln3d_oper(.true. , .false., vspl, grdo)
    stat = base_spln3d_oper(.false., .false., vspl, grdo)
    call bwi_fio_vmod_updt(grdo, vgrid, grdo_, "rite")
    stat = base_spln3d_clos()
    deallocate(grdo)
    deallocate(vspl)
    stat = 0
  end function

  integer function bwi_hesnor_oper(hess_, igrid, hgrid) result(stat)
    character(len=*)                      :: hess_
    type(img_grid)                        :: igrid
    type(hes_grid)                        :: hgrid

    integer                               :: izs
    real                                  :: hess_max

    real, dimension(:,:,:,:), allocatable :: hess

    if (igrid%ax%n /=hgrid%ax_modl%n ) call seperr("ERROR: igrid%ax%n/=hgrid%ax_modl%n")
    if (igrid%ay%n /=hgrid%ay_modl%n ) call seperr("ERROR: igrid%ay%n/=hgrid%ay_modl%n")
    if (igrid%az%n /=hgrid%az_modl%n ) call seperr("ERROR: igrid%az%n/=hgrid%az_modl%n")
    if (igrid%ahx%n/=hgrid%ahx_modl%n) call seperr("ERROR: igrid%ahx%n/=hgrid%ahx_modl%n")
    if (igrid%ahy%n/=hgrid%ahy_modl%n) call seperr("ERROR: igrid%ahy%n/=hgrid%ahy_modl%n") 
    allocate(hess(hgrid%ahx_modl%n,hgrid%ahy_modl%n,hgrid%ax_modl%n,hgrid%ay_modl%n))
    hess_max = 0
    do izs=1, igrid%az%n
       call bwi_fio_imag_updt(hess, igrid, izs, hess_, "reed")
       if (hess_max<=maxval(hess)) hess_max = maxval(hess)
    end do
    do izs=1, igrid%az%n
       call bwi_fio_imag_updt(hess, igrid, izs, hess_, "reed")
       hess = hess/hess_max
       call bwi_fio_imag_updt(hess, igrid, izs, hess_, "rite")
    end do
    deallocate(hess) 
    stat = 0
  end function

end module
