!!.. This module contains various kinds of imaging conditions
!!----------------------------------------------------------------
!!.. Author: Yaxun Tang, tang@sep.stanford.edu
!!----------------------------------------------------------------
!! This version includes the adjoint of imaging condition
!!----------------------------------------------------------------
module bwi_hescpt_mod
  use  bwi_struct_mod
  implicit none


  integer                         , private :: indx, indy, indz, nw, nsh
  type(out_grid)                  , private :: ogrid
  type(pro_grid)                  , private :: pgrid 
  integer, dimension(:,:), pointer, private :: index_mxs, index_mxr
  integer, dimension(:,:), pointer, private :: index_mys, index_myr
  integer, dimension(:,:), pointer, private :: index_mzs, index_mzr
  integer, external               , private :: sreed_window, srite_window


contains

  integer function bwi_hescpt_init(ogrid_in, pgrid_in, index_mxs_in, index_mxr_in, index_mys_in, index_myr_in, hestyp) result(stat)
    type(out_grid)                  :: ogrid_in
    type(pro_grid)                  :: pgrid_in
    integer, dimension(:,:), target :: index_mxs_in, index_mxr_in
    integer, dimension(:,:), target :: index_mys_in, index_myr_in
    integer                         :: hestyp

    integer                         :: imx, imy, imz, jmx, jmy
    integer                         :: ihx, ihy, ihz
    integer                         :: jhx, jhy, jhz
    real                            :: hx, hy, hz

    stat = 1

    indx = ogrid_in%index_outpro_xmin
    indy = ogrid_in%index_outpro_ymin
    indz = ogrid_in%index_outpro_zmin

    ogrid = ogrid_in
    pgrid = pgrid_in

    index_mxs => index_mxs_in
    index_mxr => index_mxr_in
    index_mys => index_mys_in
    index_myr => index_myr_in 

!   indeses for the source
!   ----------------------
    do ihx=1, ogrid%ahx%n
       hx  = ogrid%ahx%o+(ihx-1)*ogrid%ahx%d
       jhx = nint(hx/pgrid%ax%d)
       do imx=1, ogrid%ax%n
          jmx = ogrid%image_jx*(imx-1)+1
          index_mxs(ihx,imx) = max(1, min(indx+jmx-1-jhx,pgrid%ax%n))
       end do 
    end do

    do ihy=1, ogrid%ahy%n
       hy  = ogrid%ahy%o+(ihy-1)*ogrid%ahy%d
       jhy = nint(hy/pgrid%ay%d)
       do imy=1, ogrid%ay%n
          jmy = ogrid%image_jy*(imy-1)+1
          index_mys(ihy,imy) = max(1, min(indy+jmy-1-jhy,pgrid%ay%n))
       end do
    end do

!   indeses for the receiver
!   ------------------------
    if (hestyp==2) then
       do ihx=1, ogrid%ahx%n
          hx  = ogrid%ahx%o+(ihx-1)*ogrid%ahx%d
          jhx = nint(hx/pgrid%ax%d)
          do imx=1, ogrid%ax%n
             jmx = ogrid%image_jx*(imx-1)+1
             index_mxr(ihx,imx) = max(1, min(indx+jmx-1+jhx,pgrid%ax%n))
          end do
       end do

       do ihy=1, ogrid%ahy%n
          hy  = ogrid%ahy%o+(ihy-1)*ogrid%ahy%d
          jhy = nint(hy/pgrid%ay%d)
          do imy=1, ogrid%ay%n
             jmy = ogrid%image_jy*(imy-1)+1
             index_myr(ihy,imy) = max(1, min(indy+jmy-1+jhy,pgrid%ay%n))
          end do
       end do
    end if

    stat = 0

  end function

  integer function bwi_hescpt_clos() result(stat)
    stat = 1

    indx = 0; indy = 0; indz = 0

    nullify(index_mxs, index_mxr)
    nullify(index_mys, index_myr)

    stat = 0
  end function

  integer function bwi_souint_oper(hess, wfds_slice, ws) result(stat)
    real   , dimension(:,:,:,:) :: hess
    complex, dimension(:,:)     :: wfds_slice
    real                        :: ws

    integer                     :: imx, imy, ihx, ihy
    integer                     :: jxs, jys


    do imy=1, ogrid%ay%n
    do imx=1, ogrid%ax%n   
    do ihy=1, ogrid%ahy%n 
    do ihx=1, ogrid%ahx%n
       jxs = index_mxs(ihx,imx)
       jys = index_mys(ihy,imy)
       hess(ihx,ihy,imx,imy) = hess(ihx,ihy,imx,imy)                             &
                             + wfds_slice(jxs,jys)*conjg(wfds_slice(jxs,jys))    &
                             * ws**4
    end do
    end do
    end do
    end do

    stat = 0
  end function
!---------------------------------------------------------------------------------------------------
  integer function bwi_hesenc_oper(hess, wfds_slice, wfdr_slice, sfnc, ws) result(stat)
    real   , dimension(:,:,:,:) :: hess
    complex, dimension(:,:)     :: wfds_slice !!.. wfds(nx, ny, nz)
    complex, dimension(:,:)     :: wfdr_slice
    complex                     :: sfnc
    real                        :: ws

    integer                     :: imx, imy, ihx, ihy
    integer                     :: jxs, jys, jxr, jyr

    do imy=1, ogrid%ay%n
    do imx=1, ogrid%ax%n
    do ihy=1, ogrid%ahy%n
    do ihx=1, ogrid%ahx%n
       jxs = index_mxs(ihx,imx); jys = index_mys(ihy,imy)
       jxr = index_mxr(ihx,imx); jyr = index_myr(ihy,imy)
       hess(ihx,ihy,imx,imy) = hess(ihx,ihy,imx,imy)                             &
                             + wfds_slice(jxs,jys)*conjg(wfds_slice(jxs,jys))    &
                             * wfdr_slice(jxr,jyr)*conjg(wfdr_slice(jxr,jyr))    & 
                             / (sfnc*conjg(sfnc)+1.e-6)*(ws**4)
    end do
    end do
    end do
    end do
    stat = 0
  end function


!======================================================================================================================

end module
