!!$
!!$=head1 NAME
!!$
!!$ Lloyd_vel - Select reference velocities 
!!$
!!$=head1 SYNOPSIS
!!$
!!$Lloyd_vel  < vel.H >out.H
!!$
!!$
!!$=over 4
!!$
!!$
!!$=item nref (integer)
!!$
!!$  Number of reference velocities
!!$
!!$=item vel_map (local)
!!$
!!$  [.false.] Whether or not stdout is the reference velocities
!!$
!!$            or substiture selected reference velocity in orignal velocity model
!!$
!!$=item min_region_pct (real)
!!$
!!$  [1.]  Minimum percentage of the velocity field associated with a given reference
!!$
!!$=item min_dev 
!!$
!!$  [6.]  Minimum deviation (in percent) between reference velocities
!!$
!!$=item niter_lloyd 
!!$
!!$  [30] Number of iteratons of the lloyd to run
!!$
!!$=item When doing offset gathers  ahy_n,ahy_d,ahy_o = axis 4 of D and R
!!$
!!$=item nzs must be a multiple of az__n
!!$
!!$=item nws*nmpi_threads*nomp_threads must be a multiple of aw__n
!!$
!!$=cut
!!$
!!$

program bwi_lloyd
use sep3d_struct_mod
use lloyd_mod
use sep
implicit none
type(sep3d) :: attr_s
real, allocatable :: attr(:,:),ref_attr(:,:),slice(:,:),junk(:,:),scales(:)
integer,allocatable :: iregion(:),nstart(:)
integer :: nref,i,iref,izs,ia,ib,imap,n(4), ntotal
logical :: vel_map

call sep_init(SOURCE)

call from_param("nref",nref,20)
call from_param("vel_map",vel_map,.false.)
call init_lloyd();


call init_sep3d("in",attr_s,"INPUT")

if(sep3d_ndims(attr_s)/=4) &
  call seperr("expecting 4-d cube attr,x,y,z")


n=1
n(1:sep3d_ndims(attr_s))=attr_s%n(1:sep3d_ndims(attr_s))
allocate(scales(n(1)))
allocate(nstart(n(1)))
call from_param("nstart",nstart)
scales=1
call from_param("scales",scales,scales)
if(product(nstart) > nref) &
  call seperr("product(nstart) > nref")


if(.not. vel_map) then
  call to_history("n2",n(1))
  call to_history("n1",nref)
  call to_history("n3",n(3))
end if

allocate(iregion(n(2)*n(3)),junk(n(1),n(2)*n(3)),slice(n(2)*n(3),n(1)), attr(nref,n(1)))

ntotal=0

do izs=1,n(4)
   call sreed("in",junk,size(junk)*4)
   slice=transpose(junk)
   do i=1,n(1)
     slice(:,i)=slice(:,i)*scales(i)
   end do
   iref= lloyd_go(slice, attr, nref, iregion,nstart)
   ntotal=ntotal+iref
   do i=1,n(1)
     slice(:,i)=slice(:,i)/scales(i)
     attr(:,i)=attr(:,i)/scales(i)
   end do

   if(vel_map) then
     do ib=1,size(junk,2)
       junk(:,ib)=attr(iregion(ib),:)
     end do
     call srite("out",junk,size(junk)*4)
   write(0,*) "attr", junk(2,:)
   else
     if(iref < nref) then
       attr(iref+1:,:)=-1
     end if
     call srite("out",attr,size(attr)*4)

   end if
end do
write(0,*)"*******************************"
write(0,*)"      NTOTAL = ", ntotal
write(0,*)"*******************************"
end program
