module bwi_offdso_mod
  use  bwi_struct_mod
  implicit none
contains

  integer function bwi_offdso_oper(imag, ahx, ahy, hcut, dsotyp) result(stat)
    real, dimension(:,:,:,:) :: imag
    type(axis)               :: ahx, ahy
    real                     :: hcut
    integer                  :: dsotyp

    real                     :: hxmax, hymax, hmax
    real                     :: hx, hy, hh, wt, damp
    integer                  :: ihx, ihy

    select case (dsotyp)

      case (1) !    |h|
           hxmax = max(abs(ahx%o+(ahx%n-1)*ahx%d), abs(ahx%o))
           hymax = max(abs(ahy%o+(ahy%n-1)*ahy%d), abs(ahy%o))
           hmax  = sqrt(hxmax*hxmax+hymax*hymax)
           do ihy=1, ahy%n; hy = ahy%o+(ihy-1)*ahy%d
           do ihx=1, ahx%n; hx = ahx%o+(ihx-1)*ahx%d
              hh = max(sqrt(hx*hx+hy*hy)-hcut, 0.)/hmax
              wt = hh
              imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*wt
           end do 
           end do

      case (2) ! Gaussian taper function
           hxmax = max(abs(ahx%o+(ahx%n-1)*ahx%d), abs(ahx%o))
           hymax = max(abs(ahy%o+(ahy%n-1)*ahy%d), abs(ahy%o))
           hmax  = sqrt(hxmax*hxmax+hymax*hymax)
           damp  = 6*log(10.)/((hmax-hcut)**2)
           do ihy=1, ahy%n; hy = ahy%o+(ihy-1)*ahy%d
           do ihx=1, ahx%n; hx = ahx%o+(ihx-1)*ahx%d
              hh = sqrt(hx*hx+hy*hy)
              if (hh<hcut) then
                 wt = 0.
              else
                 wt = 1-exp(-damp*(hh-hcut)**2)
              end if
              imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*wt
           end do
           end do

!      case (3) ! modified DSO, not proved yet
!           hxmax = max(abs(ahx%o+(ahx%n-1)*ahx%d), abs(ahx%o))
!           hymax = max(abs(ahy%o+(ahy%n-1)*ahy%d), abs(ahy%o))
!           hmax  = sqrt(hxmax*hxmax+hymax*hymax)
!           do ihy=1, ahy%n; hy = ahy%o+(ihy-1)*ahy%d
!           do ihx=1, ahx%n; hx = ahx%o+(ihx-1)*ahx%d
!              hh = max(sqrt(hx*hx+hy*hy)-hcut, 0.)/hmax
!              imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*hh
!           end do
!           end do
!           do ihx=1, ahx%n/2
!              imag(ihx        ,1,:,:) = imag(ihx+1    ,1,:,:) - imag(ihx        ,1,:,:)
!              imag(ahx%n-ihx+1,1,:,:) = imag(ahx%n-ihx,1,:,:) - imag(ahx%n-ihx+1,1,:,:)
!           end do

      case default
           write(0,*) "dsotyp = [1]/2 "
           STOP
    end select
    stat = 0
  end function

end module
