module bwi_poldiv_mod2
  use  base_hlxtyp_mod
  use  base_codtrs_mod
  implicit none

  type(helix), dimension(:,:)  , pointer, private :: dipfac
  real       , dimension(:,:)  , pointer, private :: dipfac0
  integer    , dimension(:,:,:), pointer, private :: selx, sely
  integer                               , private :: nx, ny, nz, nd   
  real       , dimension(:), allocatable, private :: tt
contains


  integer function bwi_poldiv_init(dipfac_in, dipfac0_in, selx_in, sely_in, nx_in, ny_in, nz_in) result(stat)
    type(helix), dimension(:,:)  , target :: dipfac_in
    real       , dimension(:,:)  , target :: dipfac0_in
    integer    , dimension(:,:,:), target :: selx_in, sely_in
    integer                               :: nx_in, ny_in, nz_in

    dipfac  => dipfac_in 
    dipfac0 => dipfac0_in
    selx    => selx_in
    sely    => sely_in
    nx      =  nx_in
    ny      =  ny_in
    nz      =  nz_in

    nd = nx*ny*nz
    if (.not.allocated(tt)) then
       allocate(tt(nd))
    end if 
    stat = 0
  end function

  integer function bwi_poldiv_clos() result(stat)
    deallocate(tt)
    stat = 0 
  end function

  integer function bwi_poldiv_oper(adj, add, xx, yy) result(stat)
    logical                        :: adj, add 
    real, dimension(:,:,:)         :: xx, yy
    integer                        :: ix, iy, ia

    integer                        :: iselx, isely
    integer, dimension(3)          :: nn, ii, jj
    integer, dimension(:), pointer :: lag
    real   , dimension(:), pointer :: flt
    real                           :: flt0, ss

    nn = (/nx,ny,nz/)

    if (.not.add) then
       if (adj) then
          xx = 0.
       else
          yy = 0.
       end if 
    end if

    tt = 0.
    if (adj) then
       do iy=1, nd
          call line2cart(nn, iy, ii)
          tt(iy) = yy(ii(1),ii(2),ii(3))
       end do
       do iy=nd, 1, -1
          call line2cart(nn, iy, ii)
          iselx = selx(ii(1),ii(2),ii(3))
          isely = sely(ii(1),ii(2),ii(3))
          lag => dipfac(iselx,isely)%lag
          flt => dipfac(iselx,isely)%flt
          do ia=1, size(lag)
             ix = iy - lag(ia)
             if (ix<1) cycle
             tt(ix) = tt(ix) - flt(ia)*tt(iy)
          end do
       end do
       do iy=1, nd
          call line2cart(nn, iy, ii)
          iselx = selx(ii(1),ii(2),ii(3))
          isely = sely(ii(1),ii(2),ii(3))
          flt0  = dipfac0(iselx,isely)
          xx(ii(1),ii(2),ii(3)) = xx(ii(1),ii(2),ii(3)) + tt(iy)/flt0
       end do

    else
      do iy=1, nd
         call line2cart(nn, iy, ii)
         iselx = selx(ii(1),ii(2),ii(3))
         isely = sely(ii(1),ii(2),ii(3))
         flt0  = dipfac0(iselx,isely)
         tt(iy) = xx(ii(1),ii(2),ii(3))/flt0
      end do
      do iy=1, nd
         call line2cart(nn, iy, ii)
         iselx = selx(ii(1),ii(2),ii(3))
         isely = sely(ii(1),ii(2),ii(3))
         lag  => dipfac(iselx,isely)%lag
         flt  => dipfac(iselx,isely)%flt
         do ia =1, size(lag)
            ix = iy - lag(ia)
            if (ix<1) cycle
            tt(iy) = tt(iy) - flt(ia)*tt(ix)
         end do
      end do
      do iy=1, nd
         call line2cart(nn, iy, ii)
         yy(ii(1),ii(2),ii(3)) = yy(ii(1),ii(2),ii(3)) + tt(iy)
      end do

    end if
    stat = 0
  end function

end module
