! encode random sequences with different random code
module bwi_rdmgen_mod
  use base_struct_mod
  use base_fftwv3_mod
  use base_window_mod
  use sep
  implicit none

  complex, dimension(:), pointer    , private :: sfnc

  type(axis)                        , private :: aws
  integer                           , private :: ntr
  integer                           , private :: rdmtyp

contains

  integer function bwi_rdmgen_init(sfnc_in, aws_in, ntr_in, rdmtyp_in) result(stat)
    complex, dimension(:), target :: sfnc_in

    type(axis)                    :: aws_in
    integer                       :: ntr_in
    integer                       :: rdmtyp_in

    sfnc   => sfnc_in
    aws    =  aws_in
    ntr    =  ntr_in
    rdmtyp =  rdmtyp_in !!.. 1, no encoding
                        !!.. 2. phase encoding
                        !!.. 3. binary encoding
                        !!.. 4. random time delay

!    write(0,*)"........................................"
!    select case (rdmtyp)
!      case (1)
!           write(0,*)"NO ENCODING"
!      case (2)
!           write(0,*)"RANDOM PHASE ENCODING"
!      case (3)
!           write(0,*)"RANDOM BINARY ENCODING"
!      case (4)
!           write(0,*)"RANDOM TIME DELAY"
!    end select
!    write(0,*)"........................................"

    stat = 0
  end function 

  integer function bwi_rdmgen_clos() result(stat)
    nullify(sfnc)
    stat = 0
  end function

  integer function bwi_rdmgen_oper(crec) result(stat)
    complex, dimension(:,:) :: crec

    integer                 :: itr, iws
    real                    :: phase, rd, ws, tmax, PI

    PI = 3.14159265359

    crec = 0.
    call random_seed()

!   no encoding
!   -----------
    if (rdmtyp==1) then
       do itr=1, ntr
          crec(itr,:) = sfnc
       end do

!   phase code
!   ----------
    else if (rdmtyp==2) then
       do itr=1, ntr
          do iws=1, aws%n
             call random_number(phase)
             phase = 2.*PI*phase
             crec(itr,iws) = sfnc(iws)*cmplx(cos(phase),sin(phase))
          end do
       end do

!   binary code
!   -----------
    else if (rdmtyp==3) then
       do itr=1, ntr
          call random_number(rd)
          if (rd < 0.5) then
             rd = -1
          else
             rd = +1
          end if
          crec(itr,:) = sfnc*rd
       end do

!   random time delay
!   -----------------
    else if (rdmtyp==4) then
       tmax = 2.
       do itr=1, ntr
          call random_number(rd)
          rd = rd*tmax
          do iws=1, aws%n
             ws = aws%o+(iws-1)*aws%d
             phase = 2*PI*ws*rd
             crec(itr,iws) = sfnc(iws)*cmplx(cos(phase),sin(phase))
          end do
       end do

    end if
    stat = 0
  end function 

end module
