module bwi_restom_mod
  use base_struct_mod
  use bwi_fileio_mod
  use bwi_offdso_mod
  implicit none


contains

  integer function bwi_restom_oper(bimg_, hess_, imsk_, igrid, hgrid, fobj, hcut, dsotyp, dsorms, objtyp, objeps, grdfix) result(stat)
    character(len=*) :: bimg_, hess_, imsk_
    type(img_grid)   :: igrid
    type(hes_grid)   :: hgrid
    double precision :: fobj
    real             :: hcut
    integer          :: dsotyp
    logical          :: dsorms
    character(len=6) :: objtyp
    real             :: objeps
    logical          :: grdfix

    integer          :: ix, iy, izs, ihx, ihy, jhx0, jhy0
    integer          :: nsmp_offdso, nsmp_rmsdso, nsmp_imgpow
    double precision :: fobj_offdso, fobj_imgpow

    real, dimension(:,:,:,:), allocatable :: imag    
    real, dimension(:,:)    , allocatable :: imsk, istk 
    real, dimension(:,:)    , allocatable :: rms_img, rms_dso
    real, dimension(:,:,:,:), allocatable :: temp, htmp

    allocate(imag(igrid%ahx%n,igrid%ahy%n,igrid%ax%n,igrid%ay%n))
    allocate(imsk(igrid%ax%n,igrid%ay%n))

    if (hgrid%wnthes) allocate(htmp(igrid%ahx%n,igrid%ahy%n,igrid%ax%n,igrid%ay%n))

!    nsmp_offdso = igrid%ax%n*igrid%ay%n*igrid%az%n*igrid%ahx%n*igrid%ahy%n
!    nsmp_rmsdso = igrid%ax%n*igrid%ay%n 
!    nsmp_imgpow = igrid%ax%n*igrid%ay%n*igrid%az%n

nsmp_offdso = 1
nsmp_rmsdso = 1
nsmp_imgpow = 1

    select case (objtyp)

      case ("imgpow")
           fobj = 0.
           do izs=1, igrid%az%n
              call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "reed")
              if (igrid%imgmsk) then
                 call bwi_fio_imsk_updt(imsk, igrid, izs, imsk_, "reed")
                 do ihy=1, igrid%ahy%n
                 do ihx=1, igrid%ahx%n
                    imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*imsk(:,:)
                 end do
                 end do
              end if
              if (hgrid%wnthes) then
                 call bwi_fio_imag_updt(htmp, igrid, izs, hess_, "reed")
                 fobj = fobj - sum(imag*imag/(htmp+1.e-5))/nsmp_imgpow
                 imag = -imag/sqrt(htmp+1.e-5)/nsmp_imgpow
              else
                 fobj = fobj - 0.5*sum(imag*imag)/nsmp_imgpow
                 imag = -imag/nsmp_imgpow
              end if
              call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "rite")
           end do
!           write(0,*)"imgpow, fobj = ", fobj
 
      case ("offdso")

           if (dsorms) then
              allocate(rms_img(igrid%ax%n,igrid%ay%n))
              allocate(rms_dso(igrid%ax%n,igrid%ay%n))
              allocate(temp(igrid%ahx%n,igrid%ahy%n,igrid%ax%n,igrid%ay%n))
!             compute the RMS weight
!             ----------------------
              rms_img = 0.
              rms_dso = 0.
              do izs=1, igrid%az%n
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "reed")
                 if (igrid%imgmsk) then
                    call bwi_fio_imsk_updt(imsk, igrid, izs, imsk_, "reed")
                    do ihy=1, igrid%ahy%n
                    do ihx=1, igrid%ahx%n
                       imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*imsk(:,:)
                    end do
                    end do
                 end if
                 temp = imag
                 stat = bwi_offdso_oper(temp, igrid%ahx, igrid%ahy, hcut, dsotyp)
                 do ihy=1, igrid%ahy%n
                 do ihx=1, igrid%ahx%n
                    rms_img = rms_img + imag(ihx,ihy,:,:)**2
                    rms_dso = rms_dso + temp(ihx,ihy,:,:)**2
                 end do
                 end do
              end do
!             calculate the objective function
!             --------------------------------
              fobj = 0.5*sum(rms_dso/(rms_img+epsilon(hcut)))/nsmp_rmsdso
!             compute the residual and the objective function
!             ------------------------------------------------
              do izs=1, igrid%az%n
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "reed")
                 if (igrid%imgmsk) then
                    call bwi_fio_imsk_updt(imsk, igrid, izs, imsk_, "reed")
                    do ihy=1, igrid%ahy%n
                    do ihx=1, igrid%ahx%n
                       imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*imsk(:,:)
                    end do
                    end do
                 end if
                 temp = imag
!                calculate DI
!                -------------
                 stat = bwi_offdso_oper(temp, igrid%ahx, igrid%ahy, hcut, dsotyp)
!                calculate D'DI
!                --------------
                 stat = bwi_offdso_oper(temp, igrid%ahx, igrid%ahy, hcut, dsotyp)
!                image residual
!                -------------- 
                 do ihy=1, igrid%ahy%n
                 do ihx=1, igrid%ahx%n
                    imag(ihx,ihy,:,:) = (temp(ihx,ihy,:,:)*rms_img - imag(ihx,ihy,:,:)*rms_dso) &
                                      / (rms_img+epsilon(hcut))/(rms_img+epsilon(hcut))/nsmp_rmsdso
                 end do
                 end do
!                only for inline offsets
!                -----------------------
                 if (grdfix) then
                    jhx0 = 1.5 + (0-igrid%ahx%o)/igrid%ahx%d
                    jhy0 = 1.5 + (0-igrid%ahy%o)/igrid%ahy%d
                    do ihx=1, igrid%ahx%n/2
                       imag(ihx              ,1,:,:) = imag(ihx+1          ,1,:,:) - imag(ihx              ,1,:,:)
                       imag(igrid%ahx%n-ihx+1,1,:,:) = imag(igrid%ahx%n-ihx,1,:,:) - imag(igrid%ahx%n-ihx+1,1,:,:)
                    end do
                    imag(jhx0,jhy0,:,:) = 0.
                 end if
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "rite")
              end do
              deallocate(rms_dso, rms_img, temp)
!              write(0,*)"normalized offdso, fobj = ", fobj 

           else

              fobj = 0.
              do izs=1, igrid%az%n
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "reed")
                 if (igrid%imgmsk) then
                    call bwi_fio_imsk_updt(imsk, igrid, izs, imsk_, "reed")
                    do ihy=1, igrid%ahy%n
                    do ihx=1, igrid%ahx%n
                       imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*imsk(:,:)
                    end do
                    end do
                 end if
!                calculate  DI
!                -------------
                 stat = bwi_offdso_oper(imag, igrid%ahx, igrid%ahy, hcut, dsotyp)
!                evaluate the objective function
!                ------------------------------- 
                 fobj = fobj + 0.5*sum(imag*imag)/nsmp_offdso
!                calculate  D'DI
!                ---------------
                 stat = bwi_offdso_oper(imag, igrid%ahx, igrid%ahy, hcut, dsotyp)
                 imag = imag/nsmp_offdso
!                only for inline offsets
!                -----------------------
                 if (grdfix) then
                    do ihx=1, igrid%ahx%n/2
                       imag(ihx              ,1,:,:) = imag(ihx+1          ,1,:,:) - imag(ihx              ,1,:,:)
                       imag(igrid%ahx%n-ihx+1,1,:,:) = imag(igrid%ahx%n-ihx,1,:,:) - imag(igrid%ahx%n-ihx+1,1,:,:)
                    end do
                 end if
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "rite")
              end do
!              write(0,*)"offdso, fobj = ", fobj
           end if

      case ("combos")

           if (dsorms) then
              allocate(rms_img(igrid%ax%n,igrid%ay%n))
              allocate(rms_dso(igrid%ax%n,igrid%ay%n))
              allocate(istk(igrid%ax%n,igrid%ay%n))
              allocate(temp(igrid%ahx%n,igrid%ahy%n,igrid%ax%n,igrid%ay%n))
!             compute the RMS weight
!             ----------------------
              rms_img = 0.
              rms_dso = 0.
              do izs=1, igrid%az%n
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "reed")
                 if (igrid%imgmsk) then
                    call bwi_fio_imsk_updt(imsk, igrid, izs, imsk_, "reed")
                    do ihy=1, igrid%ahy%n
                    do ihx=1, igrid%ahx%n
                       imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*imsk(:,:)
                    end do
                    end do
                 end if
                 temp = imag
                 stat = bwi_offdso_oper(temp, igrid%ahx, igrid%ahy, hcut, dsotyp)
                 do ihy=1, igrid%ahy%n
                 do ihx=1, igrid%ahx%n
                    rms_img = rms_img + imag(ihx,ihy,:,:)**2
                    rms_dso = rms_dso + temp(ihx,ihy,:,:)**2
                 end do
                 end do
              end do

!             compute the residual and the objective function
!             ------------------------------------------------
              fobj_offdso = sum(rms_dso/(rms_img+epsilon(hcut)))/nsmp_rmsdso
              fobj_imgpow = 0.
              jhx0 = 1.5 + (0-igrid%ahx%o)/igrid%ahx%d
              jhy0 = 1.5 + (0-igrid%ahy%o)/igrid%ahy%d
              do izs=1, igrid%az%n
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "reed")
                 if (igrid%imgmsk) then
                    call bwi_fio_imsk_updt(imsk, igrid, izs, imsk_, "reed")
                    do ihy=1, igrid%ahy%n
                    do ihx=1, igrid%ahx%n
                       imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*imsk(:,:)
                    end do
                    end do
                 end if
                 temp = imag
                 istk = imag(jhx0,jhy0,:,:)
!                calculate DI
                 stat = bwi_offdso_oper(temp, igrid%ahx, igrid%ahy, hcut, dsotyp)
!                calculate D'DI
                 stat = bwi_offdso_oper(temp, igrid%ahx, igrid%ahy, hcut, dsotyp)
!                evaluate the objective function
                 fobj_imgpow = fobj_imgpow - sum(istk*istk)/nsmp_imgpow
!                calculate the image residual
                 do ihy=1, igrid%ahy%n
                 do ihx=1, igrid%ahx%n
                    imag(ihx,ihy,:,:) = (temp(ihx,ihy,:,:)*rms_img - imag(ihx,ihy,:,:)*rms_dso) &
                                      / (rms_img+epsilon(hcut))/(rms_img+epsilon(hcut))/nsmp_rmsdso
                 end do
                 end do
                 imag(jhx0,jhy0,:,:) = imag(jhx0,jhy0,:,:) - istk*objeps/nsmp_imgpow
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "rite")
              end do
              fobj = 0.5*fobj_offdso + 0.5*fobj_imgpow*objeps
              deallocate(rms_dso, rms_img, istk, temp)

           else

              allocate(istk(igrid%ax%n,igrid%ay%n))
              fobj_offdso = 0.
              fobj_imgpow = 0.
              jhx0 = 1.5 + (0-igrid%ahx%o)/igrid%ahx%d
              jhy0 = 1.5 + (0-igrid%ahy%o)/igrid%ahy%d
              do izs=1, igrid%az%n
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "reed")
                 if (igrid%imgmsk) then
                    call bwi_fio_imsk_updt(imsk, igrid, izs, imsk_, "reed")
                    do ihy=1, igrid%ahy%n
                    do ihx=1, igrid%ahx%n
                       imag(ihx,ihy,:,:) = imag(ihx,ihy,:,:)*imsk(:,:)
                    end do
                    end do
                 end if
                 istk = imag(jhx0,jhy0,:,:)
!                calculate DI
                 stat = bwi_offdso_oper(imag, igrid%ahx, igrid%ahy, hcut, dsotyp)
!                evaluate the objective functions
                 fobj_offdso = fobj_offdso + sum(imag*imag)/nsmp_offdso
                 fobj_imgpow = fobj_imgpow - sum(istk*istk)/nsmp_imgpow
!                calculate D'DI
                 stat = bwi_offdso_oper(imag, igrid%ahx, igrid%ahy, hcut, dsotyp)
                 imag = imag/nsmp_offdso
!                calculate the image residual
                 imag(jhx0,jhy0,:,:) = imag(jhx0,jhy0,:,:) - istk*objeps/nsmp_imgpow
                 call bwi_fio_imag_updt(imag, igrid, izs, bimg_, "rite")
              end do
              fobj = 0.5*fobj_offdso + 0.5*fobj_imgpow*objeps
              deallocate(istk)
           end if

!           write(0,*) "fobj_offdso = ", fobj_offdso*0.5, "fobj_imgpow = ", fobj_imgpow*0.5, "fobj = ", fobj

    end select
    deallocate(imag)
    deallocate(imsk)
    if (hgrid%wnthes) deallocate(htmp)
    stat = 0
  end function

end module
