!!! search direction is flipped?????????
module bwi_search_mod
  use base_struct_mod
  use bwi_kernel_mod
  use bwi_qrestom_mod
  implicit none

contains

  integer function bwi_sdstep_oper(srch_, grdo_, qgrid) result(stat)
    character(len=*)                    :: srch_, grdo_
    type(q_grid)                        :: qgrid
    real, dimension(:,:,:), allocatable :: srch

    allocate(srch(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))
    call bwi_fio_qmod_updt(srch, qgrid, grdo_, "reed")
    srch = -srch
    call bwi_fio_qmod_updt(srch, qgrid, srch_, "rite")
    deallocate(srch)
    stat = 0.
  end function

  integer function bwi_cgstep_oper(srch_, grdo_, grdp_, qgrid, restart) result(stat)
    character(len=*)                    :: srch_, grdo_, grdp_
    type(q_grid)                        :: qgrid
    logical                             :: restart

    double precision                    :: beta
    real, dimension(:,:,:), allocatable :: grdo
    real, dimension(:,:,:), allocatable :: grdp
    real, dimension(:,:,:), allocatable :: srch

    allocate(grdo(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))
    allocate(grdp(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))
    allocate(srch(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))

!   gradient for the k'th iteration
    call bwi_fio_qmod_updt(grdo, qgrid, grdo_, "reed") ! k
!   gradient for the k+1'th iteration
    call bwi_fio_qmod_updt(grdp, qgrid, grdp_, "reed") ! k+1
!   restarting
    if (restart) then
       beta = 0.
    else
!      compute beta, using the Polak-Ribiere method
       beta = sum(dprod(grdp,grdp-grdo))/(sum(dprod(grdo,grdo))+epsilon(beta))
!      safe guarding
       beta = max(beta, 0.)
    end if
!    write(0,*)"beta = ", beta
!   read the search direction for the k'th iteration (current)
    call bwi_fio_qmod_updt(srch, qgrid, srch_, "reed")
!   update the search direction
    srch = - grdo + beta*srch
!   write out new search direction
    call bwi_fio_qmod_updt(srch, qgrid, srch_, "rite")
!   shuffle the gradient
    call bwi_fio_qmod_updt(grdp, qgrid, grdo_, "rite")
    deallocate(grdo)
    deallocate(grdp)
    deallocate(srch)
    stat = 0
  end function
 
  integer function bwi_linsec_oper(csou_, crec_, bvel_, bq_, bimg_, hess_, srch_, imsk_, vgrid, qgrid,  igrid, hgrid, nsrc, alph, fobj, &
                                   objtyp,qcomp, qcomp_clip, nwin, lwin, fmin, fmax, lsniter, neigh, ver, ng, nref, nsmth, slop_neg) result(stat)

 


    character(len=*)                    :: csou_, crec_, bvel_,bq_, bimg_, hess_, srch_, imsk_
    type(vel_grid)                      :: vgrid
    type(q_grid)                        :: qgrid
    type(img_grid)                      :: igrid
    type(hes_grid)                      :: hgrid
    integer                             :: nsrc
    real                                :: alph
    double precision                    :: fobj
    character(len=6)                    :: objtyp


    double precision                    :: f(nsrc), fopt, norm
    real                                :: sfac
    integer                             :: iopt, isrc

    real, dimension(:,:,:), allocatable :: grdo
    real, dimension(:,:,:), allocatable :: grdp
    real, dimension(:,:,:), allocatable :: srch

  logical  				          :: qcomp
  real	 					  :: qcomp_clip
  logical					  :: neigh, ver, slop_neg
  integer					  :: ng, nref
  integer					  :: nwin, lwin, lsniter, nsmth
  real						  :: fmin, fmax
  integer					  :: i1, i2, i3
    allocate(grdo(qgrid%ax%n,qgrid%ay%n,qgrid%az%n)) 
    allocate(grdp(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))
    allocate(srch(qgrid%ax%n,qgrid%ay%n,qgrid%az%n))

    sfac = 0.5*(nsrc+1)

!   read the background model to be updated
!   ---------------------------------------
    call bwi_fio_qmod_updt(grdo, qgrid, bq_, "reed")
!   read the search direction
!   -------------------------
    call bwi_fio_qmod_updt(srch, qgrid, srch_, "reed")
!   normalize
!   ---------
    norm = maxval(abs(srch))
    srch = srch/norm


!   perturb the model
!   -----------------
    do isrc=1, nsrc
!      trial velocity model
!      -------------------- 
       grdp = grdo + alph*isrc*srch
    do i1=1, qgrid%ax%n
    do i2=1, qgrid%ay%n
    do i3=1, qgrid%az%n
	if (grdp(i1, i2, i3) .gt. 5.) then
		grdp(i1, i2, i3)=5.
	else if (grdp(i1, i2, i3) .lt. 1.3) then
		grdp(i1, i2, i3)=1.3
	end if
    end do
    end do
    end do

    do i1=1, qgrid%ax%n
    do i2=1, qgrid%ay%n
    do i3=1, 50
	grdp(i1, i2, i3)=5.
    end do
    end do
    end do


       call bwi_fio_qmod_updt(grdp, qgrid, bq_, "rite")
!      compute the image using the trial velocity
!      ------------------------------------------
	write(0,*) "grdp",minval(grdp),maxval(grdp)

       stat = bwi_kerimg_oper(.true., .false., qcomp, csou_, crec_, bvel_, bq_, bimg_, qcomp_clip)

!      compute the objective function
!      ------------------------------
       stat = bwi_qrestom_oper(bimg_, hess_, imsk_, igrid, hgrid, f(isrc), objtyp,nwin, lwin, fmin, fmax, lsniter, neigh, ver, ng, nref, nsmth, slop_neg )



    end do
!   find the minimum one
!   --------------------
    fopt = f(1)
    iopt = 1
    do isrc=1, nsrc
	write(0,*) "isrc,nsrc,f(isrc)", isrc,nsrc, f(isrc)
       if (f(isrc)<=fopt) then
          iopt = isrc
          fopt = f(isrc)
	write(0,*) "iopt, fopt", iopt, fopt
       end if
    end do
!   compare with the current objective function value
!   -------------------------------------------------
!        write(0,*) "fopt, fobj", fopt, fobj
! 
!        grdp = grdo + alph*iopt*srch
!        alph = alph*iopt/sfac
!        fobj = fopt
!        write(0,*) "fobj", fobj
!!!??? comment if
   write(0,*) "fobj1", fobj
   if (f(iopt)<=1.2*fobj) then

	write(0,*) " f(iopt)", f(iopt)
      grdp = grdo + alph*iopt*srch
      alph = alph*iopt/sfac
      fobj = f(iopt)
      write(0,*) "fobj2", fobj
   else
      grdp = grdo
      alph = alph * 0.5
   end if
    do i1=1, qgrid%ax%n
    do i2=1, qgrid%ay%n
    do i3=1, qgrid%az%n
	if (grdp(i1, i2, i3) .gt. 5.0) then
		grdp(i1, i2, i3)=5.0
	else if (grdp(i1, i2, i3) .lt. 1.3) then
		grdp(i1, i2, i3)=1.3
	end if
    end do
    end do
    end do

  !  do i1=1, qgrid%ax%n
   ! do i2=1, qgrid%ay%n
    !do i3=1, 50
	!grdp(i1, i2, i3)=5.
    !end do
   !end do
   ! end do

	write(0,*) "grdp", minval(grdp), maxval(grdp)
	write(0,*) "fobj", fobj
    call bwi_fio_qmod_updt(grdp, qgrid, bq_, "rite")
    deallocate(grdo)
    deallocate(grdp)
    deallocate(srch)
    stat = 0
  end function


end module
