program main
  use sep
  use base_hlxtyp_mod
  use base_poldiv_mod
  use base_wilson_mod
  implicit none

  type(helix) :: dipfac, dipflt
  real        :: PI
  integer     :: na, ia, n1, ndip, niter, idip, stat
  real        :: da, oa, odip, ddip, u1, u2, m, p, eps, dip, dipflt0, dipfac0 
  real        :: dip_min, dip_max

  PI = 3.14159265359

  call initpar()

  call from_param("n1"   , n1   , 101 )
  call from_param("na"   , na   , 1   )
  call from_param("niter", niter, 20  )
  call from_param("eps"  , eps  , 0.01)

  call from_param("dip_min", dip_min, -50.)
  call from_param("dip_max", dip_max,  50.)
  call from_param("ndip"   , ndip   , 1   )

  oa = 0.; da = 1.

  if (ndip==1) then
     odip = dip_min
     ddip = 1.
  else
     odip = dip_min
     ddip = (dip_max-dip_min)/(ndip-1)
  end if

  call to_history("n1", 3*na); call to_history("d1", da  ); call to_history("o1", oa  )
  call to_history("n2", ndip); call to_history("d2", ddip); call to_history("o2", odip)

  call allocatehelix(dipflt, 4   )
  call allocatehelix(dipfac, 3*na)

  dipflt%lag = (/1, n1-1, n1, n1+1/)

  do ia=1, na
     dipfac%lag(ia)      = ia
     dipfac%lag(na+ia)   = n1+ia-na
     dipfac%lag(2*na+ia) = n1+ia
  end do

  stat = base_wilson_init(maxval(dipflt%lag)+1)

  write(0,*)"na    = ", na
  write(0,*)"n1    = ", n1
  write(0,*)"niter = ", niter
  write(0,*)"eps   = ", eps

  do idip=1, ndip
     dip = odip+(idip-1)*ddip
     u1  =  cos(dip*PI/180.)
     u2  = -sin(dip*PI/180.)
     m   = (u1-u2)*0.5
     p   = (u1+u2)*0.5

     dipflt0    = 1. + eps
     dipflt%flt = (/2*m*p, -p*p, -2*m*p, -m*m/)
     dipfac0    = 0.
     dipfac%flt = 0.
     stat = base_wilson_oper(niter, dipflt0, dipflt, dipfac0, dipfac)
     call sep_write(dipfac%flt)
  end do
  stat = base_wilson_clos()
  call deallocatehelix(dipflt)
  call deallocatehelix(dipfac)
  call exit()
end program
