program main
  use base_ioutil_mod
  use bwi_diphlx_mod
  use bwi_grdprc_mod
  implicit none

  type(axis)         :: axk, ayk, azk, adip
  type(vel_grid)     :: vgrid
  type(img_grid)     :: igrid
  character(len=5)   :: scheme
  character(len=256) :: grdo_, gmsk_, bimg_, dipx_, dipy_

  logical            :: smtgrd, wntdip, gscale
  character(len=6)   :: smttyp
  integer            :: rectx_grd, recty_grd, rectz_grd, nsmt_grd, stat
  integer            :: rectx_dip, recty_dip, rectz_dip, nsmt_dip
  integer            :: ndip, dipitr, na, idip
  real               :: xk_min, yk_min, zk_min
  real               :: xk_max, yk_max, zk_max 
  real               :: dip_min, dip_max, dipeps 

  integer    , dimension(:,:,:), allocatable :: selx, sely
  type(helix), dimension(:)    , allocatable :: dfcx, dfcy
  real       , dimension(:)    , allocatable :: dfcx0, dfcy0
  real       , dimension(:,:,:), allocatable :: dipx, dipy           

  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  grdo_ = "grdo"
  gmsk_ = "gmsk"
  bimg_ = "bimg"
  dipx_ = "dipx"
  dipy_ = "dipy"

  call from_param("smtgrd", smtgrd      , .true.  )
  call from_param("smttyp", smttyp      , "triflt")
  call from_param("grdmsk", vgrid%grdmsk, .false. )
  call from_param("gscale", gscale      , .true.  )

  call auxinout(grdo_)
  if (vgrid%grdmsk) call auxinout(gmsk_)

  call hinaxis(grdo_, vgrid%ax, 1)
  call hinaxis(grdo_, vgrid%ay, 2)
  call hinaxis(grdo_, vgrid%az, 3)

! apply gradient smoothing
! ------------------------
  if (smtgrd) then
     select case (smttyp)
       case ("dipflt")
            call from_param("rectx_dip", rectx_dip, 1       )
            call from_param("recty_dip", recty_dip, 1       )
            call from_param("rectz_dip", rectz_dip, 1       )
            call from_param("nsmt_dip" , nsmt_dip , 1       )
            call from_param("dip_min"  , dip_min  , 0.      )
            call from_param("dip_max"  , dip_max  , 0.      )
            call from_param("ndip"     , adip%n   , 1       )
            call from_param("dipitr"   , dipitr   , 100     )
            call from_param("dipeps"   , dipeps   , 0.01    )
            call from_param("na"       , na       , 5       )
            call from_param("scheme"   , scheme   , "UW2x3" )
            call from_param("wntdip"   , wntdip   , .false. )
            if (adip%n==1) then
               adip%o = dip_min
               adip%d = 1.
            else
               adip%o = dip_min
               adip%d = (dip_max-dip_min)/(adip%n-1)
            end if
            adip%n = adip%n + 1

            call auxinout(bimg_)
            call hinaxis(bimg_, igrid%ahx, 1)
            call hinaxis(bimg_, igrid%ahy, 2)
            call hinaxis(bimg_, igrid%ax , 3)
            call hinaxis(bimg_, igrid%ay , 4)
            call hinaxis(bimg_, igrid%az , 5)

            vgrid%index_imgvel_xmin = 1.5 + (igrid%ax%o-vgrid%ax%o)/vgrid%ax%d
            vgrid%index_imgvel_ymin = 1.5 + (igrid%ay%o-vgrid%ay%o)/vgrid%ay%d
            vgrid%index_imgvel_zmin = 1.5 + (igrid%az%o-vgrid%az%o)/vgrid%az%d
            vgrid%index_imgvel_xmax = 1.5 + (igrid%ax%o+(igrid%ax%n-1)*igrid%ax%d-vgrid%ax%o)/vgrid%ax%d
            vgrid%index_imgvel_ymax = 1.5 + (igrid%ay%o+(igrid%ay%n-1)*igrid%ay%d-vgrid%ay%o)/vgrid%ay%d
            vgrid%index_imgvel_zmax = 1.5 + (igrid%az%o+(igrid%az%n-1)*igrid%az%d-vgrid%az%o)/vgrid%az%d

!           tabulate dip filters in x
!           -------------------------
            if (vgrid%ax%n>1) then
               allocate(dfcx(adip%n), dfcx0(adip%n))
               allocate(dipx(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               allocate(selx(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               do idip=1, adip%n
                  call allocatehelix(dfcx(idip), 3*na)
               end do
               stat = bwi_diphlx_init(adip, dipeps, vgrid%ax%d, vgrid%az%d, vgrid%ax%n, na, dipitr, scheme)
               stat = bwi_diphlx_oper(dfcx, dfcx0)
               stat = bwi_diphlx_clos()
            end if

!           tabulate dip filters in y
!           -------------------------
            if (vgrid%ay%n>1) then
               allocate(dfcy(adip%n), dfcy0(adip%n))
               allocate(dipy(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               allocate(sely(vgrid%ax%n,vgrid%ay%n,vgrid%az%n))
               do idip=1, adip%n
                  call allocatehelix(dfcy(idip), 3*na)
               end do
               stat = bwi_diphlx_init(adip, dipeps, vgrid%ay%d, vgrid%az%d, vgrid%ay%n, na, dipitr, scheme)
               stat = bwi_diphlx_oper(dfcy, dfcy0)
               stat = bwi_diphlx_clos()
            end if

!           estimate dips
!           -------------
            stat = bwi_update_dips(bimg_, dipx, dipy, selx, sely, igrid, vgrid, adip, rectx_dip, recty_dip, rectz_dip, nsmt_dip)

!           output dips
!           -----------
            if (wntdip) then
               if (vgrid%ax%n>1) then
                  call auxinout(dipx_)
                  call houaxis(dipx_, vgrid%ax, 1)
                  call houaxis(dipx_, vgrid%ay, 2)
                  call houaxis(dipx_, vgrid%az, 3)
                  call sseek(dipx_, 0, 0)
                  call sep_write(dipx, dipx_)
               end if
               if (vgrid%ay%n>1) then
                  call auxinout(dipy_)
                  call houaxis(dipy_, vgrid%ax, 1)
                  call houaxis(dipy_, vgrid%ay, 2)
                  call houaxis(dipy_, vgrid%az, 3)
                  call sseek(dipy_, 0, 0)
                  call sep_write(dipy, dipy_)
               end if
            end if

!           apply dip filtering
!           -------------------
            stat = bwi_dipflt_oper(grdo_, dfcx, dfcy, dfcx0, dfcy0, selx, sely, vgrid)

            stat = bwi_triflt_oper(grdo_, vgrid, min(vgrid%ax%n,2), min(vgrid%ay%n,2), min(vgrid%az%n,2), 2)

       case ("triflt")
            call from_param("rectx_grd", rectx_grd, 1)
            call from_param("recty_grd", recty_grd, 1)
            call from_param("rectz_grd", rectz_grd, 1)
            call from_param("nsmt_grd" , nsmt_grd , 1)
            stat = bwi_triflt_oper(grdo_, vgrid, rectx_grd, recty_grd, rectz_grd, nsmt_grd)

       case ("spline")
            call from_param("xk_min",xk_min,0.); call from_param("xk_max",xk_max,0.); call from_param("nxk",axk%n,1)
            call from_param("yk_min",yk_min,0.); call from_param("yk_max",yk_max,0.); call from_param("nyk",ayk%n,1)
            call from_param("zk_min",zk_min,0.); call from_param("zk_max",zk_max,0.); call from_param("nzk",azk%n,1)
            xk_min = vgrid%ax%o + (int(1.5+(xk_min-vgrid%ax%o)/vgrid%ax%d)-1)*vgrid%ax%d
            xk_max = vgrid%ax%o + (int(1.5+(xk_max-vgrid%ax%o)/vgrid%ax%d)-1)*vgrid%ax%d
            yk_min = vgrid%ay%o + (int(1.5+(yk_min-vgrid%ay%o)/vgrid%ay%d)-1)*vgrid%ay%d
            yk_max = vgrid%ay%o + (int(1.5+(yk_max-vgrid%ay%o)/vgrid%ay%d)-1)*vgrid%ay%d
            zk_min = vgrid%az%o + (int(1.5+(zk_min-vgrid%az%o)/vgrid%az%d)-1)*vgrid%az%d
            zk_max = vgrid%az%o + (int(1.5+(zk_max-vgrid%az%o)/vgrid%az%d)-1)*vgrid%az%d
            axk%o = xk_min; axk%d = 1.
            ayk%o = yk_min; ayk%d = 1.
            azk%o = zk_min; azk%d = 1.
            if (axk%n>1) then
               axk%o = xk_min; axk%d = (xk_max-xk_min)/(axk%n-1)
            end if
            if (ayk%n>1) then
               ayk%o = yk_min; ayk%d = (yk_max-yk_min)/(ayk%n-1)
            end if
            if (azk%n>1) then
               azk%o = zk_min; azk%d = (zk_max-zk_min)/(azk%n-1)
            end if
            stat = bwi_spline_oper(grdo_, vgrid, axk, ayk, azk)

     end select
  end if

! scale and normalize the gradient
! --------------------------------
  if (gscale) then
     stat = bwi_gscale_oper(grdo_, vgrid)
  end if

! apply gradient mask
! ------------------- 
  if (vgrid%grdmsk) then
     stat = bwi_grdmsk_oper(grdo_, gmsk_, vgrid)
  end if

  call sep_end_prog()

  call exit()

end program
