module off2ang_fileio_mod
  implicit none

  integer, external, private :: srite_window, sreed_window

  interface off2ang_fio_gtho
    module procedure off2ang_fio_gtho_real
    module procedure off2ang_fio_gtho_cplx
  end interface

  interface off2ang_fio_gtha
    module procedure off2ang_fio_gtha_real
    module procedure off2ang_fio_gtha_cplx
  end interface

contains
  
  subroutine off2ang_fio_gtho_real(gtho,nx,ny,nz,nhx,nhy,jx,jy,filetag,flag)
    real, dimension(:,:,:) :: gtho
    integer                :: nx, ny, nz, nhx, nhy, jx, jy
    character(len=*)       :: filetag
    character(len=4)       :: flag 
    integer, dimension(5)  :: n_cube, j_wind, n_wind, f_wind
    integer                :: n_dim_cube, err

    n_dim_cube = 5
    n_cube     = (/nhx,nhy,nz,nx,ny/) 
    j_wind     = (/1,1,1,1,1/)
    n_wind     = (/nhx,nhy,nz,1,1/)
    f_wind     = (/0,0,0,jx-1,jy-1/)

    if (flag=="rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,gtho)
    else if (flag=="reed") then
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,gtho)
    end if
    return
  end subroutine

  subroutine off2ang_fio_gtho_cplx(gtho,nx,ny,nz,nhx,nhy,jx,jy,filetag,flag)
    complex, dimension(:,:,:) :: gtho
    integer                   :: nx, ny, nz, nhx, nhy, jx, jy
    character(len=*)          :: filetag
    character(len=4)          :: flag
    integer, dimension(5)     :: n_cube, j_wind, n_wind, f_wind
    integer                   :: n_dim_cube, err

    n_dim_cube = 5
    n_cube     = (/nhx,nhy,nz,nx,ny/)
    j_wind     = (/1,1,1,1,1/)
    n_wind     = (/nhx,nhy,nz,1,1/)
    f_wind     = (/0,0,0,jx-1,jy-1/)

    if (flag=="rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,gtho)
    else if (flag=="reed") then
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,gtho)
    end if
    return
  end subroutine

  subroutine off2ang_fio_gtha_real(gtha,nx,ny,nz,nagl_rfl,nagl_azm,jx,jy,filetag,flag)
    real, dimension(:,:,:) :: gtha
    integer                :: nx, ny, nz, nagl_rfl, nagl_azm, jx, jy
    character(len=*)       :: filetag
    character(len=4)       :: flag
    integer, dimension(5)  :: n_cube, j_wind, n_wind, f_wind
    integer                :: n_dim_cube, err

    n_dim_cube = 5
    n_cube     = (/nagl_rfl,nagl_azm,nz,nx,ny/)
    j_wind     = (/1,1,1,1,1/)
    n_wind     = (/nagl_rfl,nagl_azm,nz,1,1/)
    f_wind     = (/0,0,0,jx-1,jy-1/)

    if (flag=="rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,gtha)
    else if (flag=="reed") then
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,gtha)
    end if
    return
  end subroutine

  subroutine off2ang_fio_gtha_cplx(gtha,nx,ny,nz,nagl_rfl,nagl_azm,jx,jy,filetag,flag)
    complex, dimension(:,:,:) :: gtha
    integer                   :: nx, ny, nz, nagl_rfl, nagl_azm, jx, jy
    character(len=*)          :: filetag
    character(len=4)          :: flag
    integer, dimension(5)     :: n_cube, j_wind, n_wind, f_wind
    integer                   :: n_dim_cube, err

    n_dim_cube = 5
    n_cube     = (/nagl_rfl,nagl_azm,nz,nx,ny/)
    j_wind     = (/1,1,1,1,1/)
    n_wind     = (/nagl_rfl,nagl_azm,nz,1,1/)
    f_wind     = (/0,0,0,jx-1,jy-1/)

    if (flag=="rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,gtha)
    else if (flag=="reed") then
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,gtha)
    end if
    return
  end subroutine



end module
