module off2ang_kernsd_mod
  use  base_struct_mod
  implicit none

  type(axis)     , private :: ahx, ahy, az
  type(axis)     , private :: aagl_rfl, aagl_azm
  real, parameter, private :: PI = 3.14159265359

  interface off2ang_kernsd_oper
    module procedure off2ang_kersdc_oper
    module procedure off2ang_kersdr_oper
  end interface

contains

  integer function off2ang_kernsd_init(ahx_in,ahy_in,aagl_rfl_in,aagl_azm_in,az_in) result(stat)
    type(axis) :: ahx_in, ahy_in
    type(axis) :: aagl_rfl_in, aagl_azm_in, az_in

    ahx      = ahx_in
    ahy      = ahy_in
    aagl_rfl = aagl_rfl_in
    aagl_azm = aagl_azm_in
    az       = az_in

    stat = 0
  end function

  integer function off2ang_kernsd_clos() result(stat)
    stat = 0
  end function

  integer function off2ang_kersdc_oper(gtha, gtho) result(stat)
    complex, dimension(:,:,:) :: gtha, gtho

    integer                   :: iagl_rfl, iz, ihx, jzm, jzp, jz
    real                      :: agl_rfl, z, hx, zh, rzo, fzm, fzp 

    stat = 1
    gtha = 0.
    do iagl_rfl=1, aagl_rfl%n; agl_rfl = aagl_rfl%o+(iagl_rfl-1)*aagl_rfl%d
    do iz      =1, az%n      ; z       = az%o+(iz-1)*az%d
    do ihx     =1, ahx%n     ; hx      = ahx%o+(ihx-1)*ahx%d

       zh = z + tan(agl_rfl*PI/180.)*hx
!             jz = 1.5+(zh-ogrid%az%o)/ogrid%az%d
!             jz = max(1, jz)
!             jz = min(jz, ogrid%az%n)
!             gtha(iagl_rfl,1,iz) = gtha(iagl_rfl,1,iz) + gtho(ihx,1,jz) 
       rzo = 1.+(zh-az%o)/az%d
       jzm = rzo
       if (jzm<1 .or. jzm>az%n-1) cycle
       jzp = jzm+1
       fzp = rzo-jzm
       fzm = 1-fzp
       gtha(iagl_rfl,1,iz) = gtha(iagl_rfl,1,iz)+gtho(ihx,1,jzm)*fzm+gtho(ihx,1,jzp)*fzp

    end do 
    end do
    end do
    stat = 0
  end function

  integer function off2ang_kersdr_oper(gtha, gtho) result(stat)
    real   , dimension(:,:,:) :: gtha, gtho

    integer                   :: iagl_rfl, iz, ihx, jzm, jzp, jz
    real                      :: agl_rfl, z, hx, zh, rzo, fzm, fzp

    stat = 1
    gtha = 0.
    do iagl_rfl=1, aagl_rfl%n; agl_rfl = aagl_rfl%o+(iagl_rfl-1)*aagl_rfl%d
    do iz      =1, az%n      ; z       = az%o+(iz-1)*az%d
    do ihx     =1, ahx%n     ; hx      = ahx%o+(ihx-1)*ahx%d

       zh = z + tan(agl_rfl*PI/180.)*hx

       rzo = 1.+(zh-az%o)/az%d
       jzm = rzo
       if (jzm<1 .or. jzm>az%n-1) cycle 
       jzp = jzm+1
       fzp = rzo-jzm
       fzm = 1-fzp
       gtha(iagl_rfl,1,iz) = gtha(iagl_rfl,1,iz)+gtho(ihx,1,jzm)*fzm+gtho(ihx,1,jzp)*fzp

    end do
    end do
    end do
    stat = 0
  end function




end module
